# -*- coding: utf-8 -*-

# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from functools import partial
import apiclient
import ciso8601
import datetime
import json
import logging
import mock
import multiprocessing
import os
import pwd
import random
import re
import select
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import unittest
import uuid

import arvados
import arvados.commands.put as arv_put
from . import arvados_testutil as tutil

from .arvados_testutil import ArvadosBaseTestCase, fake_httplib2_response
from . import run_test_server

class ArvadosPutResumeCacheTest(ArvadosBaseTestCase):
    CACHE_ARGSET = [
        [],
        ['/dev/null'],
        ['/dev/null', '--filename', 'empty'],
        ['/tmp']
        ]

    def tearDown(self):
        super(ArvadosPutResumeCacheTest, self).tearDown()
        try:
            self.last_cache.destroy()
        except AttributeError:
            pass

    def cache_path_from_arglist(self, arglist):
        return arv_put.ResumeCache.make_path(arv_put.parse_arguments(arglist))

    def test_cache_names_stable(self):
        for argset in self.CACHE_ARGSET:
            self.assertEqual(self.cache_path_from_arglist(argset),
                              self.cache_path_from_arglist(argset),
                              "cache name changed for {}".format(argset))

    def test_cache_names_unique(self):
        results = []
        for argset in self.CACHE_ARGSET:
            path = self.cache_path_from_arglist(argset)
            self.assertNotIn(path, results)
            results.append(path)

    def test_cache_names_simple(self):
        # The goal here is to make sure the filename doesn't use characters
        # reserved by the filesystem.  Feel free to adjust this regexp as
        # long as it still does that.
        bad_chars = re.compile(r'[^-\.\w]')
        for argset in self.CACHE_ARGSET:
            path = self.cache_path_from_arglist(argset)
            self.assertFalse(bad_chars.search(os.path.basename(path)),
                             "path too exotic: {}".format(path))

    def test_cache_names_ignore_argument_order(self):
        self.assertEqual(
            self.cache_path_from_arglist(['a', 'b', 'c']),
            self.cache_path_from_arglist(['c', 'a', 'b']))
        self.assertEqual(
            self.cache_path_from_arglist(['-', '--filename', 'stdin']),
            self.cache_path_from_arglist(['--filename', 'stdin', '-']))

    def test_cache_names_differ_for_similar_paths(self):
        # This test needs names at / that don't exist on the real filesystem.
        self.assertNotEqual(
            self.cache_path_from_arglist(['/_arvputtest1', '/_arvputtest2']),
            self.cache_path_from_arglist(['/_arvputtest1/_arvputtest2']))

    def test_cache_names_ignore_irrelevant_arguments(self):
        # Workaround: parse_arguments bails on --filename with a directory.
        path1 = self.cache_path_from_arglist(['/tmp'])
        args = arv_put.parse_arguments(['/tmp'])
        args.filename = 'tmp'
        path2 = arv_put.ResumeCache.make_path(args)
        self.assertEqual(path1, path2,
                         "cache path considered --filename for directory")
        self.assertEqual(
            self.cache_path_from_arglist(['-']),
            self.cache_path_from_arglist(['-', '--max-manifest-depth', '1']),
            "cache path considered --max-manifest-depth for file")

    def test_cache_names_treat_negative_manifest_depths_identically(self):
        base_args = ['/tmp', '--max-manifest-depth']
        self.assertEqual(
            self.cache_path_from_arglist(base_args + ['-1']),
            self.cache_path_from_arglist(base_args + ['-2']))

    def test_cache_names_treat_stdin_consistently(self):
        self.assertEqual(
            self.cache_path_from_arglist(['-', '--filename', 'test']),
            self.cache_path_from_arglist(['/dev/stdin', '--filename', 'test']))

    def test_cache_names_identical_for_synonymous_names(self):
        self.assertEqual(
            self.cache_path_from_arglist(['.']),
            self.cache_path_from_arglist([os.path.realpath('.')]))
        testdir = self.make_tmpdir()
        looplink = os.path.join(testdir, 'loop')
        os.symlink(testdir, looplink)
        self.assertEqual(
            self.cache_path_from_arglist([testdir]),
            self.cache_path_from_arglist([looplink]))

    def test_cache_names_different_by_api_host(self):
        config = arvados.config.settings()
        orig_host = config.get('ARVADOS_API_HOST')
        try:
            name1 = self.cache_path_from_arglist(['.'])
            config['ARVADOS_API_HOST'] = 'x' + (orig_host or 'localhost')
            self.assertNotEqual(name1, self.cache_path_from_arglist(['.']))
        finally:
            if orig_host is None:
                del config['ARVADOS_API_HOST']
            else:
                config['ARVADOS_API_HOST'] = orig_host

    @mock.patch('arvados.keep.KeepClient.head')
    def test_resume_cache_with_current_stream_locators(self, keep_client_head):
        keep_client_head.side_effect = [True]
        thing = {}
        thing['_current_stream_locators'] = ['098f6bcd4621d373cade4e832627b4f6+4', '1f253c60a2306e0ee12fb6ce0c587904+6']
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
        self.last_cache.save(thing)
        self.last_cache.close()
        resume_cache = arv_put.ResumeCache(self.last_cache.filename)
        self.assertNotEqual(None, resume_cache)

    @mock.patch('arvados.keep.KeepClient.head')
    def test_resume_cache_with_finished_streams(self, keep_client_head):
        keep_client_head.side_effect = [True]
        thing = {}
        thing['_finished_streams'] = [['.', ['098f6bcd4621d373cade4e832627b4f6+4', '1f253c60a2306e0ee12fb6ce0c587904+6']]]
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
        self.last_cache.save(thing)
        self.last_cache.close()
        resume_cache = arv_put.ResumeCache(self.last_cache.filename)
        self.assertNotEqual(None, resume_cache)

    @mock.patch('arvados.keep.KeepClient.head')
    def test_resume_cache_with_finished_streams_error_on_head(self, keep_client_head):
        keep_client_head.side_effect = Exception('Locator not found')
        thing = {}
        thing['_finished_streams'] = [['.', ['098f6bcd4621d373cade4e832627b4f6+4', '1f253c60a2306e0ee12fb6ce0c587904+6']]]
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
        self.last_cache.save(thing)
        self.last_cache.close()
        resume_cache = arv_put.ResumeCache(self.last_cache.filename)
        self.assertNotEqual(None, resume_cache)
        resume_cache.check_cache()

    def test_basic_cache_storage(self):
        thing = ['test', 'list']
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
        self.last_cache.save(thing)
        self.assertEqual(thing, self.last_cache.load())

    def test_empty_cache(self):
        with tempfile.NamedTemporaryFile() as cachefile:
            cache = arv_put.ResumeCache(cachefile.name)
        self.assertRaises(ValueError, cache.load)

    def test_cache_persistent(self):
        thing = ['test', 'list']
        path = os.path.join(self.make_tmpdir(), 'cache')
        cache = arv_put.ResumeCache(path)
        cache.save(thing)
        cache.close()
        self.last_cache = arv_put.ResumeCache(path)
        self.assertEqual(thing, self.last_cache.load())

    def test_multiple_cache_writes(self):
        thing = ['short', 'list']
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
        # Start writing an object longer than the one we test, to make
        # sure the cache file gets truncated.
        self.last_cache.save(['long', 'long', 'list'])
        self.last_cache.save(thing)
        self.assertEqual(thing, self.last_cache.load())

    def test_cache_is_locked(self):
        with tempfile.NamedTemporaryFile() as cachefile:
            _ = arv_put.ResumeCache(cachefile.name)
            self.assertRaises(arv_put.ResumeCacheConflict,
                              arv_put.ResumeCache, cachefile.name)

    def test_cache_stays_locked(self):
        with tempfile.NamedTemporaryFile() as cachefile:
            self.last_cache = arv_put.ResumeCache(cachefile.name)
            path = cachefile.name
        self.last_cache.save('test')
        self.assertRaises(arv_put.ResumeCacheConflict,
                          arv_put.ResumeCache, path)

    def test_destroy_cache(self):
        cachefile = tempfile.NamedTemporaryFile(delete=False)
        try:
            cache = arv_put.ResumeCache(cachefile.name)
            cache.save('test')
            cache.destroy()
            try:
                arv_put.ResumeCache(cachefile.name)
            except arv_put.ResumeCacheConflict:
                self.fail("could not load cache after destroying it")
            self.assertRaises(ValueError, cache.load)
        finally:
            if os.path.exists(cachefile.name):
                os.unlink(cachefile.name)

    def test_restart_cache(self):
        path = os.path.join(self.make_tmpdir(), 'cache')
        cache = arv_put.ResumeCache(path)
        cache.save('test')
        cache.restart()
        self.assertRaises(ValueError, cache.load)
        self.assertRaises(arv_put.ResumeCacheConflict,
                          arv_put.ResumeCache, path)


class ArvPutUploadJobTest(run_test_server.TestCaseWithServers,
                          ArvadosBaseTestCase):

    def setUp(self):
        super(ArvPutUploadJobTest, self).setUp()
        run_test_server.authorize_with('active')
        # Temp files creation
        self.tempdir = tempfile.mkdtemp()
        subdir = os.path.join(self.tempdir, 'subdir')
        os.mkdir(subdir)
        data = "x" * 1024 # 1 KB
        for i in range(1, 5):
            with open(os.path.join(self.tempdir, str(i)), 'w') as f:
                f.write(data * i)
        with open(os.path.join(subdir, 'otherfile'), 'w') as f:
            f.write(data * 5)
        # Large temp file for resume test
        _, self.large_file_name = tempfile.mkstemp()
        fileobj = open(self.large_file_name, 'w')
        # Make sure to write just a little more than one block
        for _ in range((arvados.config.KEEP_BLOCK_SIZE>>20)+1):
            data = random.choice(['x', 'y', 'z']) * 1024 * 1024 # 1 MiB
            fileobj.write(data)
        fileobj.close()
        # Temp dir containing small files to be repacked
        self.small_files_dir = tempfile.mkdtemp()
        data = 'y' * 1024 * 1024 # 1 MB
        for i in range(1, 70):
            with open(os.path.join(self.small_files_dir, str(i)), 'w') as f:
                f.write(data + str(i))
        self.arvfile_write = getattr(arvados.arvfile.ArvadosFileWriter, 'write')
        # Temp dir to hold a symlink to other temp dir
        self.tempdir_with_symlink = tempfile.mkdtemp()
        os.symlink(self.tempdir, os.path.join(self.tempdir_with_symlink, 'linkeddir'))
        os.symlink(os.path.join(self.tempdir, '1'),
                   os.path.join(self.tempdir_with_symlink, 'linkedfile'))

    def tearDown(self):
        super(ArvPutUploadJobTest, self).tearDown()
        shutil.rmtree(self.tempdir)
        os.unlink(self.large_file_name)
        shutil.rmtree(self.small_files_dir)
        shutil.rmtree(self.tempdir_with_symlink)

    def test_non_regular_files_are_ignored_except_symlinks_to_dirs(self):
        def pfunc(x):
            with open(x, 'w') as f:
                f.write('test')
        fifo_filename = 'fifo-file'
        fifo_path = os.path.join(self.tempdir_with_symlink, fifo_filename)
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkeddir')))
        os.mkfifo(fifo_path)
        producer = multiprocessing.Process(target=pfunc, args=(fifo_path,))
        producer.start()
        cwriter = arv_put.ArvPutUploadJob([self.tempdir_with_symlink])
        cwriter.start(save_collection=False)
        if producer.exitcode is None:
            # If the producer is still running, kill it. This should always be
            # before any assertion that may fail.
            producer.terminate()
            producer.join(1)
        self.assertIn('linkeddir', cwriter.manifest_text())
        self.assertNotIn(fifo_filename, cwriter.manifest_text())

    def test_symlinks_are_followed_by_default(self):
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkeddir')))
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkedfile')))
        cwriter = arv_put.ArvPutUploadJob([self.tempdir_with_symlink])
        cwriter.start(save_collection=False)
        self.assertIn('linkeddir', cwriter.manifest_text())
        self.assertIn('linkedfile', cwriter.manifest_text())
        cwriter.destroy_cache()

    def test_symlinks_are_not_followed_when_requested(self):
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkeddir')))
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkedfile')))
        cwriter = arv_put.ArvPutUploadJob([self.tempdir_with_symlink],
                                          follow_links=False)
        cwriter.start(save_collection=False)
        self.assertNotIn('linkeddir', cwriter.manifest_text())
        self.assertNotIn('linkedfile', cwriter.manifest_text())
        cwriter.destroy_cache()
        # Check for bug #17800: passed symlinks should also be ignored.
        linked_dir = os.path.join(self.tempdir_with_symlink, 'linkeddir')
        cwriter = arv_put.ArvPutUploadJob([linked_dir], follow_links=False)
        cwriter.start(save_collection=False)
        self.assertNotIn('linkeddir', cwriter.manifest_text())
        cwriter.destroy_cache()

    def test_no_empty_collection_saved(self):
        self.assertTrue(os.path.islink(os.path.join(self.tempdir_with_symlink, 'linkeddir')))
        linked_dir = os.path.join(self.tempdir_with_symlink, 'linkeddir')
        cwriter = arv_put.ArvPutUploadJob([linked_dir], follow_links=False)
        cwriter.start(save_collection=True)
        self.assertIsNone(cwriter.manifest_locator())
        self.assertEqual('', cwriter.manifest_text())
        cwriter.destroy_cache()

    def test_passing_nonexistant_path_raise_exception(self):
        uuid_str = str(uuid.uuid4())
        with self.assertRaises(arv_put.PathDoesNotExistError):
            arv_put.ArvPutUploadJob(["/this/path/does/not/exist/{}".format(uuid_str)])

    def test_writer_works_without_cache(self):
        cwriter = arv_put.ArvPutUploadJob(['/dev/null'], resume=False)
        cwriter.start(save_collection=False)
        self.assertEqual(". d41d8cd98f00b204e9800998ecf8427e+0 0:0:null\n", cwriter.manifest_text())

    def test_writer_works_with_cache(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b'foo')
            f.flush()
            cwriter = arv_put.ArvPutUploadJob([f.name])
            cwriter.start(save_collection=False)
            self.assertEqual(0, cwriter.bytes_skipped)
            self.assertEqual(3, cwriter.bytes_written)
            # Don't destroy the cache, and start another upload
            cwriter_new = arv_put.ArvPutUploadJob([f.name])
            cwriter_new.start(save_collection=False)
            cwriter_new.destroy_cache()
            self.assertEqual(3, cwriter_new.bytes_skipped)
            self.assertEqual(3, cwriter_new.bytes_written)

    def make_progress_tester(self):
        progression = []
        def record_func(written, expected):
            progression.append((written, expected))
        return progression, record_func

    def test_progress_reporting(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(b'foo')
            f.flush()
            for expect_count in (None, 8):
                progression, reporter = self.make_progress_tester()
                cwriter = arv_put.ArvPutUploadJob([f.name],
                                                  reporter=reporter)
                cwriter.bytes_expected = expect_count
                cwriter.start(save_collection=False)
                cwriter.destroy_cache()
                self.assertIn((3, expect_count), progression)

    def test_writer_upload_directory(self):
        cwriter = arv_put.ArvPutUploadJob([self.tempdir])
        cwriter.start(save_collection=False)
        cwriter.destroy_cache()
        self.assertEqual(1024*(1+2+3+4+5), cwriter.bytes_written)

    def test_resume_large_file_upload(self):
        def wrapped_write(*args, **kwargs):
            data = args[1]
            # Exit only on last block
            if len(data) < arvados.config.KEEP_BLOCK_SIZE:
                # Simulate a checkpoint before quitting. Ensure block commit.
                self.writer._update(final=True)
                raise SystemExit("Simulated error")
            return self.arvfile_write(*args, **kwargs)

        with mock.patch('arvados.arvfile.ArvadosFileWriter.write',
                        autospec=True) as mocked_write:
            mocked_write.side_effect = wrapped_write
            writer = arv_put.ArvPutUploadJob([self.large_file_name],
                                             replication_desired=1)
            # We'll be accessing from inside the wrapper
            self.writer = writer
            with self.assertRaises(SystemExit):
                writer.start(save_collection=False)
            # Confirm that the file was partially uploaded
            self.assertGreater(writer.bytes_written, 0)
            self.assertLess(writer.bytes_written,
                            os.path.getsize(self.large_file_name))
        # Retry the upload
        writer2 = arv_put.ArvPutUploadJob([self.large_file_name],
                                          replication_desired=1)
        writer2.start(save_collection=False)
        self.assertEqual(writer.bytes_written + writer2.bytes_written - writer2.bytes_skipped,
                         os.path.getsize(self.large_file_name))
        writer2.destroy_cache()
        del(self.writer)

    # Test for bug #11002
    def test_graceful_exit_while_repacking_small_blocks(self):
        def wrapped_commit(*args, **kwargs):
            raise SystemExit("Simulated error")

        with mock.patch('arvados.arvfile._BlockManager.commit_bufferblock',
                        autospec=True) as mocked_commit:
            mocked_commit.side_effect = wrapped_commit
            # Upload a little more than 1 block, wrapped_commit will make the first block
            # commit to fail.
            # arv-put should not exit with an exception by trying to commit the collection
            # as it's in an inconsistent state.
            writer = arv_put.ArvPutUploadJob([self.small_files_dir],
                                             replication_desired=1)
            try:
                with self.assertRaises(SystemExit):
                    writer.start(save_collection=False)
            except arvados.arvfile.UnownedBlockError:
                self.fail("arv-put command is trying to use a corrupted BlockManager. See https://dev.arvados.org/issues/11002")
        writer.destroy_cache()

    def test_no_resume_when_asked(self):
        def wrapped_write(*args, **kwargs):
            data = args[1]
            # Exit only on last block
            if len(data) < arvados.config.KEEP_BLOCK_SIZE:
                # Simulate a checkpoint before quitting.
                self.writer._update()
                raise SystemExit("Simulated error")
            return self.arvfile_write(*args, **kwargs)

        with mock.patch('arvados.arvfile.ArvadosFileWriter.write',
                        autospec=True) as mocked_write:
            mocked_write.side_effect = wrapped_write
            writer = arv_put.ArvPutUploadJob([self.large_file_name],
                                             replication_desired=1)
            # We'll be accessing from inside the wrapper
            self.writer = writer
            with self.assertRaises(SystemExit):
                writer.start(save_collection=False)
            # Confirm that the file was partially uploaded
            self.assertGreater(writer.bytes_written, 0)
            self.assertLess(writer.bytes_written,
                            os.path.getsize(self.large_file_name))
        # Retry the upload, this time without resume
        writer2 = arv_put.ArvPutUploadJob([self.large_file_name],
                                          replication_desired=1,
                                          resume=False)
        writer2.start(save_collection=False)
        self.assertEqual(writer2.bytes_skipped, 0)
        self.assertEqual(writer2.bytes_written,
                         os.path.getsize(self.large_file_name))
        writer2.destroy_cache()
        del(self.writer)

    def test_no_resume_when_no_cache(self):
        def wrapped_write(*args, **kwargs):
            data = args[1]
            # Exit only on last block
            if len(data) < arvados.config.KEEP_BLOCK_SIZE:
                # Simulate a checkpoint before quitting.
                self.writer._update()
                raise SystemExit("Simulated error")
            return self.arvfile_write(*args, **kwargs)

        with mock.patch('arvados.arvfile.ArvadosFileWriter.write',
                        autospec=True) as mocked_write:
            mocked_write.side_effect = wrapped_write
            writer = arv_put.ArvPutUploadJob([self.large_file_name],
                                             replication_desired=1)
            # We'll be accessing from inside the wrapper
            self.writer = writer
            with self.assertRaises(SystemExit):
                writer.start(save_collection=False)
            # Confirm that the file was partially uploaded
            self.assertGreater(writer.bytes_written, 0)
            self.assertLess(writer.bytes_written,
                            os.path.getsize(self.large_file_name))
        # Retry the upload, this time without cache usage
        writer2 = arv_put.ArvPutUploadJob([self.large_file_name],
                                          replication_desired=1,
                                          resume=False,
                                          use_cache=False)
        writer2.start(save_collection=False)
        self.assertEqual(writer2.bytes_skipped, 0)
        self.assertEqual(writer2.bytes_written,
                         os.path.getsize(self.large_file_name))
        writer2.destroy_cache()
        del(self.writer)

    def test_dry_run_feature(self):
        def wrapped_write(*args, **kwargs):
            data = args[1]
            # Exit only on last block
            if len(data) < arvados.config.KEEP_BLOCK_SIZE:
                # Simulate a checkpoint before quitting.
                self.writer._update()
                raise SystemExit("Simulated error")
            return self.arvfile_write(*args, **kwargs)

        with mock.patch('arvados.arvfile.ArvadosFileWriter.write',
                        autospec=True) as mocked_write:
            mocked_write.side_effect = wrapped_write
            writer = arv_put.ArvPutUploadJob([self.large_file_name],
                                             replication_desired=1)
            # We'll be accessing from inside the wrapper
            self.writer = writer
            with self.assertRaises(SystemExit):
                writer.start(save_collection=False)
            # Confirm that the file was partially uploaded
            self.assertGreater(writer.bytes_written, 0)
            self.assertLess(writer.bytes_written,
                            os.path.getsize(self.large_file_name))
        with self.assertRaises(arv_put.ArvPutUploadIsPending):
            # Retry the upload using dry_run to check if there is a pending upload
            writer2 = arv_put.ArvPutUploadJob([self.large_file_name],
                                              replication_desired=1,
                                              dry_run=True)
        # Complete the pending upload
        writer3 = arv_put.ArvPutUploadJob([self.large_file_name],
                                          replication_desired=1)
        writer3.start(save_collection=False)
        with self.assertRaises(arv_put.ArvPutUploadNotPending):
            # Confirm there's no pending upload with dry_run=True
            writer4 = arv_put.ArvPutUploadJob([self.large_file_name],
                                              replication_desired=1,
                                              dry_run=True)
        # Test obvious cases
        with self.assertRaises(arv_put.ArvPutUploadIsPending):
            arv_put.ArvPutUploadJob([self.large_file_name],
                                    replication_desired=1,
                                    dry_run=True,
                                    resume=False,
                                    use_cache=False)
        with self.assertRaises(arv_put.ArvPutUploadIsPending):
            arv_put.ArvPutUploadJob([self.large_file_name],
                                    replication_desired=1,
                                    dry_run=True,
                                    resume=False)
        del(self.writer)

class CachedManifestValidationTest(ArvadosBaseTestCase):
    class MockedPut(arv_put.ArvPutUploadJob):
        def __init__(self, cached_manifest=None):
            self._state = arv_put.ArvPutUploadJob.EMPTY_STATE
            self._state['manifest'] = cached_manifest
            self._api_client = mock.MagicMock()
            self.logger = mock.MagicMock()
            self.num_retries = 1

    def datetime_to_hex(self, dt):
        return hex(int(time.mktime(dt.timetuple())))[2:]

    def setUp(self):
        super(CachedManifestValidationTest, self).setUp()
        self.block1 = "fdba98970961edb29f88241b9d99d890" # foo
        self.block2 = "37b51d194a7513e45b56f6524f2d51f2" # bar
        self.template = ". "+self.block1+"+3+Asignature@%s "+self.block2+"+3+Anothersignature@%s 0:3:foofile.txt 3:6:barfile.txt\n"

    def test_empty_cached_manifest_is_valid(self):
        put_mock = self.MockedPut()
        self.assertEqual(None, put_mock._state.get('manifest'))
        self.assertTrue(put_mock._cached_manifest_valid())
        put_mock._state['manifest'] = ''
        self.assertTrue(put_mock._cached_manifest_valid())

    def test_signature_cases(self):
        now = datetime.datetime.utcnow()
        yesterday = now - datetime.timedelta(days=1)
        lastweek = now - datetime.timedelta(days=7)
        tomorrow = now + datetime.timedelta(days=1)
        nextweek = now + datetime.timedelta(days=7)

        def mocked_head(blocks={}, loc=None):
            blk = loc.split('+', 1)[0]
            if blocks.get(blk):
                return True
            raise arvados.errors.KeepRequestError("mocked error - block invalid")

        # Block1_expiration, Block2_expiration, Block1_HEAD, Block2_HEAD, Expectation
        cases = [
            # All expired, reset cache - OK
            (yesterday, lastweek, False, False, True),
            (lastweek, yesterday, False, False, True),
            # All non-expired valid blocks - OK
            (tomorrow, nextweek, True, True, True),
            (nextweek, tomorrow, True, True, True),
            # All non-expired invalid blocks - Not OK
            (tomorrow, nextweek, False, False, False),
            (nextweek, tomorrow, False, False, False),
            # One non-expired valid block - OK
            (tomorrow, yesterday, True, False, True),
            (yesterday, tomorrow, False, True, True),
            # One non-expired invalid block - Not OK
            (tomorrow, yesterday, False, False, False),
            (yesterday, tomorrow, False, False, False),
        ]
        for case in cases:
            b1_expiration, b2_expiration, b1_valid, b2_valid, outcome = case
            head_responses = {
                self.block1: b1_valid,
                self.block2: b2_valid,
            }
            cached_manifest = self.template % (
                self.datetime_to_hex(b1_expiration),
                self.datetime_to_hex(b2_expiration),
            )
            arvput = self.MockedPut(cached_manifest)
            with mock.patch('arvados.collection.KeepClient.head') as head_mock:
                head_mock.side_effect = partial(mocked_head, head_responses)
                self.assertEqual(outcome, arvput._cached_manifest_valid(),
                    "Case '%s' should have produced outcome '%s'" % (case, outcome)
                )
                if b1_expiration > now or b2_expiration > now:
                    # A HEAD request should have been done
                    head_mock.assert_called_once()
                else:
                    head_mock.assert_not_called()


class ArvadosExpectedBytesTest(ArvadosBaseTestCase):
    TEST_SIZE = os.path.getsize(__file__)

    def test_expected_bytes_for_file(self):
        writer = arv_put.ArvPutUploadJob([__file__])
        self.assertEqual(self.TEST_SIZE,
                         writer.bytes_expected)

    def test_expected_bytes_for_tree(self):
        tree = self.make_tmpdir()
        shutil.copyfile(__file__, os.path.join(tree, 'one'))
        shutil.copyfile(__file__, os.path.join(tree, 'two'))

        writer = arv_put.ArvPutUploadJob([tree])
        self.assertEqual(self.TEST_SIZE * 2,
                         writer.bytes_expected)
        writer = arv_put.ArvPutUploadJob([tree, __file__])
        self.assertEqual(self.TEST_SIZE * 3,
                         writer.bytes_expected)

    def test_expected_bytes_for_device(self):
        writer = arv_put.ArvPutUploadJob(['/dev/null'], use_cache=False, resume=False)
        self.assertIsNone(writer.bytes_expected)
        writer = arv_put.ArvPutUploadJob([__file__, '/dev/null'])
        self.assertIsNone(writer.bytes_expected)


class ArvadosPutReportTest(ArvadosBaseTestCase):
    def test_machine_progress(self):
        for count, total in [(0, 1), (0, None), (1, None), (235, 9283)]:
            expect = ": {} written {} total\n".format(
                count, -1 if (total is None) else total)
            self.assertTrue(
                arv_put.machine_progress(count, total).endswith(expect))

    def test_known_human_progress(self):
        for count, total in [(0, 1), (2, 4), (45, 60)]:
            expect = '{:.1%}'.format(1.0*count/total)
            actual = arv_put.human_progress(count, total)
            self.assertTrue(actual.startswith('\r'))
            self.assertIn(expect, actual)

    def test_unknown_human_progress(self):
        for count in [1, 20, 300, 4000, 50000]:
            self.assertTrue(re.search(r'\b{}\b'.format(count),
                                      arv_put.human_progress(count, None)))


class ArvPutLogFormatterTest(ArvadosBaseTestCase):
    matcher = r'\(X-Request-Id: req-[a-z0-9]{20}\)'

    def setUp(self):
        super(ArvPutLogFormatterTest, self).setUp()
        self.stderr = tutil.StringIO()
        self.loggingHandler = logging.StreamHandler(self.stderr)
        self.loggingHandler.setFormatter(
            arv_put.ArvPutLogFormatter(arvados.util.new_request_id()))
        self.logger = logging.getLogger()
        self.logger.addHandler(self.loggingHandler)
        self.logger.setLevel(logging.DEBUG)

    def tearDown(self):
        self.logger.removeHandler(self.loggingHandler)
        self.stderr.close()
        self.stderr = None
        super(ArvPutLogFormatterTest, self).tearDown()

    def test_request_id_logged_only_once_on_error(self):
        self.logger.error('Ooops, something bad happened.')
        self.logger.error('Another bad thing just happened.')
        log_lines = self.stderr.getvalue().split('\n')[:-1]
        self.assertEqual(2, len(log_lines))
        self.assertRegex(log_lines[0], self.matcher)
        self.assertNotRegex(log_lines[1], self.matcher)

    def test_request_id_logged_only_once_on_debug(self):
        self.logger.debug('This is just a debug message.')
        self.logger.debug('Another message, move along.')
        log_lines = self.stderr.getvalue().split('\n')[:-1]
        self.assertEqual(2, len(log_lines))
        self.assertRegex(log_lines[0], self.matcher)
        self.assertNotRegex(log_lines[1], self.matcher)

    def test_request_id_not_logged_on_info(self):
        self.logger.info('This should be a useful message')
        log_lines = self.stderr.getvalue().split('\n')[:-1]
        self.assertEqual(1, len(log_lines))
        self.assertNotRegex(log_lines[0], self.matcher)

class ArvadosPutTest(run_test_server.TestCaseWithServers,
                     ArvadosBaseTestCase,
                     tutil.VersionChecker):
    MAIN_SERVER = {}
    Z_UUID = 'zzzzz-zzzzz-zzzzzzzzzzzzzzz'

    def call_main_with_args(self, args):
        self.main_stdout.seek(0, 0)
        self.main_stdout.truncate(0)
        self.main_stderr.seek(0, 0)
        self.main_stderr.truncate(0)
        return arv_put.main(args, self.main_stdout, self.main_stderr)

    def call_main_on_test_file(self, args=[]):
        with self.make_test_file() as testfile:
            path = testfile.name
            self.call_main_with_args(['--stream', '--no-progress'] + args + [path])
        self.assertTrue(
            os.path.exists(os.path.join(os.environ['KEEP_LOCAL_STORE'],
                                        '098f6bcd4621d373cade4e832627b4f6')),
            "did not find file stream in Keep store")

    def setUp(self):
        super(ArvadosPutTest, self).setUp()
        run_test_server.authorize_with('active')
        arv_put.api_client = None
        self.main_stdout = tutil.StringIO()
        self.main_stderr = tutil.StringIO()
        self.loggingHandler = logging.StreamHandler(self.main_stderr)
        self.loggingHandler.setFormatter(
            arv_put.ArvPutLogFormatter(arvados.util.new_request_id()))
        logging.getLogger().addHandler(self.loggingHandler)

    def tearDown(self):
        logging.getLogger().removeHandler(self.loggingHandler)
        for outbuf in ['main_stdout', 'main_stderr']:
            if hasattr(self, outbuf):
                getattr(self, outbuf).close()
                delattr(self, outbuf)
        super(ArvadosPutTest, self).tearDown()

    def test_version_argument(self):
        with tutil.redirected_streams(
                stdout=tutil.StringIO, stderr=tutil.StringIO) as (out, err):
            with self.assertRaises(SystemExit):
                self.call_main_with_args(['--version'])
        self.assertVersionOutput(out, err)

    def test_simple_file_put(self):
        self.call_main_on_test_file()

    def test_put_with_unwriteable_cache_dir(self):
        orig_cachedir = arv_put.ResumeCache.CACHE_DIR
        cachedir = self.make_tmpdir()
        os.chmod(cachedir, 0o0)
        arv_put.ResumeCache.CACHE_DIR = cachedir
        try:
            self.call_main_on_test_file()
        finally:
            arv_put.ResumeCache.CACHE_DIR = orig_cachedir
            os.chmod(cachedir, 0o700)

    def test_put_with_unwritable_cache_subdir(self):
        orig_cachedir = arv_put.ResumeCache.CACHE_DIR
        cachedir = self.make_tmpdir()
        os.chmod(cachedir, 0o0)
        arv_put.ResumeCache.CACHE_DIR = os.path.join(cachedir, 'cachedir')
        try:
            self.call_main_on_test_file()
        finally:
            arv_put.ResumeCache.CACHE_DIR = orig_cachedir
            os.chmod(cachedir, 0o700)

    def test_put_block_replication(self):
        self.call_main_on_test_file()
        with mock.patch('arvados.collection.KeepClient.local_store_put') as put_mock:
            put_mock.return_value = 'acbd18db4cc2f85cedef654fccc4a4d8+3'
            self.call_main_on_test_file(['--replication', '1'])
            self.call_main_on_test_file(['--replication', '4'])
            self.call_main_on_test_file(['--replication', '5'])
            self.assertEqual(
                [x[-1].get('copies') for x in put_mock.call_args_list],
                [1, 4, 5])

    def test_normalize(self):
        testfile1 = self.make_test_file()
        testfile2 = self.make_test_file()
        test_paths = [testfile1.name, testfile2.name]
        # Reverse-sort the paths, so normalization must change their order.
        test_paths.sort(reverse=True)
        self.call_main_with_args(['--stream', '--no-progress', '--normalize'] +
                                 test_paths)
        manifest = self.main_stdout.getvalue()
        # Assert the second file we specified appears first in the manifest.
        file_indices = [manifest.find(':' + os.path.basename(path))
                        for path in test_paths]
        self.assertGreater(*file_indices)

    def test_error_name_without_collection(self):
        self.assertRaises(SystemExit, self.call_main_with_args,
                          ['--name', 'test without Collection',
                           '--stream', '/dev/null'])

    def test_error_when_project_not_found(self):
        self.assertRaises(SystemExit,
                          self.call_main_with_args,
                          ['--project-uuid', self.Z_UUID])

    def test_error_bad_project_uuid(self):
        self.assertRaises(SystemExit,
                          self.call_main_with_args,
                          ['--project-uuid', self.Z_UUID, '--stream'])

    def test_error_when_excluding_absolute_path(self):
        tmpdir = self.make_tmpdir()
        self.assertRaises(SystemExit,
                          self.call_main_with_args,
                          ['--exclude', '/some/absolute/path/*',
                           tmpdir])

    def test_api_error_handling(self):
        coll_save_mock = mock.Mock(name='arv.collection.Collection().save_new()')
        coll_save_mock.side_effect = arvados.errors.ApiError(
            fake_httplib2_response(403), b'{}')
        with mock.patch('arvados.collection.Collection.save_new',
                        new=coll_save_mock):
            with self.assertRaises(SystemExit) as exc_test:
                self.call_main_with_args(['/dev/null'])
            self.assertLess(0, exc_test.exception.args[0])
            self.assertLess(0, coll_save_mock.call_count)
            self.assertEqual("", self.main_stdout.getvalue())

    def test_request_id_logging_on_error(self):
        matcher = r'\(X-Request-Id: req-[a-z0-9]{20}\)\n'
        coll_save_mock = mock.Mock(name='arv.collection.Collection().save_new()')
        coll_save_mock.side_effect = arvados.errors.ApiError(
            fake_httplib2_response(403), b'{}')
        with mock.patch('arvados.collection.Collection.save_new',
                        new=coll_save_mock):
            with self.assertRaises(SystemExit):
                self.call_main_with_args(['/dev/null'])
            self.assertRegex(
                self.main_stderr.getvalue(), matcher)


class ArvPutIntegrationTest(run_test_server.TestCaseWithServers,
                            ArvadosBaseTestCase):
    MAIN_SERVER = {}
    KEEP_SERVER = {'blob_signing': True}
    PROJECT_UUID = run_test_server.fixture('groups')['aproject']['uuid']

    @classmethod
    def setUpClass(cls):
        super(ArvPutIntegrationTest, cls).setUpClass()
        cls.ENVIRON = os.environ.copy()
        cls.ENVIRON['PYTHONPATH'] = ':'.join(sys.path)

    def datetime_to_hex(self, dt):
        return hex(int(time.mktime(dt.timetuple())))[2:]

    def setUp(self):
        super(ArvPutIntegrationTest, self).setUp()
        arv_put.api_client = None

    def authorize_with(self, token_name):
        run_test_server.authorize_with(token_name)
        for v in ["ARVADOS_API_HOST",
                  "ARVADOS_API_HOST_INSECURE",
                  "ARVADOS_API_TOKEN"]:
            self.ENVIRON[v] = arvados.config.settings()[v]
        arv_put.api_client = arvados.api('v1')

    def current_user(self):
        return arv_put.api_client.users().current().execute()

    def test_check_real_project_found(self):
        self.authorize_with('active')
        self.assertTrue(arv_put.desired_project_uuid(arv_put.api_client, self.PROJECT_UUID, 0),
                        "did not correctly find test fixture project")

    def test_check_error_finding_nonexistent_uuid(self):
        BAD_UUID = 'zzzzz-zzzzz-zzzzzzzzzzzzzzz'
        self.authorize_with('active')
        try:
            result = arv_put.desired_project_uuid(arv_put.api_client, BAD_UUID,
                                                  0)
        except ValueError as error:
            self.assertIn(BAD_UUID, str(error))
        else:
            self.assertFalse(result, "incorrectly found nonexistent project")

    def test_check_error_finding_nonexistent_project(self):
        BAD_UUID = 'zzzzz-tpzed-zzzzzzzzzzzzzzz'
        self.authorize_with('active')
        with self.assertRaises(apiclient.errors.HttpError):
            arv_put.desired_project_uuid(arv_put.api_client, BAD_UUID,
                                                  0)

    def test_short_put_from_stdin(self):
        # Have to run this as an integration test since arv-put can't
        # read from the tests' stdin.
        # arv-put usually can't stat(os.path.realpath('/dev/stdin')) in this
        # case, because the /proc entry is already gone by the time it tries.
        pipe = subprocess.Popen(
            [sys.executable, arv_put.__file__, '--stream'],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, env=self.ENVIRON)
        pipe.stdin.write(b'stdin test\xa6\n')
        pipe.stdin.close()
        deadline = time.time() + 5
        while (pipe.poll() is None) and (time.time() < deadline):
            time.sleep(.1)
        returncode = pipe.poll()
        if returncode is None:
            pipe.terminate()
            self.fail("arv-put did not PUT from stdin within 5 seconds")
        elif returncode != 0:
            sys.stdout.write(pipe.stdout.read())
            self.fail("arv-put returned exit code {}".format(returncode))
        self.assertIn('1cb671b355a0c23d5d1c61d59cdb1b2b+12',
                      pipe.stdout.read().decode())

    def test_sigint_logs_request_id(self):
        # Start arv-put, give it a chance to start up, send SIGINT,
        # and check that its output includes the X-Request-Id.
        input_stream = subprocess.Popen(
            ['sleep', '10'],
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        pipe = subprocess.Popen(
            [sys.executable, arv_put.__file__, '--stream'],
            stdin=input_stream.stdout, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, env=self.ENVIRON)
        # Wait for arv-put child process to print something (i.e., a
        # log message) so we know its signal handler is installed.
        select.select([pipe.stdout], [], [], 10)
        pipe.send_signal(signal.SIGINT)
        deadline = time.time() + 5
        while (pipe.poll() is None) and (time.time() < deadline):
            time.sleep(.1)
        returncode = pipe.poll()
        input_stream.terminate()
        if returncode is None:
            pipe.terminate()
            self.fail("arv-put did not exit within 5 seconds")
        self.assertRegex(pipe.stdout.read().decode(), r'\(X-Request-Id: req-[a-z0-9]{20}\)')

    def test_ArvPutSignedManifest(self):
        # ArvPutSignedManifest runs "arv-put foo" and then attempts to get
        # the newly created manifest from the API server, testing to confirm
        # that the block locators in the returned manifest are signed.
        self.authorize_with('active')

        # Before doing anything, demonstrate that the collection
        # we're about to create is not present in our test fixture.
        manifest_uuid = "00b4e9f40ac4dd432ef89749f1c01e74+47"
        with self.assertRaises(apiclient.errors.HttpError):
            arv_put.api_client.collections().get(
                uuid=manifest_uuid).execute()

        datadir = self.make_tmpdir()
        with open(os.path.join(datadir, "foo"), "w") as f:
            f.write("The quick brown fox jumped over the lazy dog")
        p = subprocess.Popen([sys.executable, arv_put.__file__,
                              os.path.join(datadir, 'foo')],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=self.ENVIRON)
        (_, err) = p.communicate()
        self.assertRegex(err.decode(), r'INFO: Collection saved as ')
        self.assertEqual(p.returncode, 0)

        # The manifest text stored in the API server under the same
        # manifest UUID must use signed locators.
        c = arv_put.api_client.collections().get(uuid=manifest_uuid).execute()
        self.assertRegex(
            c['manifest_text'],
            r'^\. 08a008a01d498c404b0c30852b39d3b8\+44\+A[0-9a-f]+@[0-9a-f]+ 0:44:foo\n')

        os.remove(os.path.join(datadir, "foo"))
        os.rmdir(datadir)

    def run_and_find_collection(self, text, extra_args=[]):
        self.authorize_with('active')
        pipe = subprocess.Popen(
            [sys.executable, arv_put.__file__] + extra_args,
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, env=self.ENVIRON)
        stdout, stderr = pipe.communicate(text.encode())
        self.assertRegex(stderr.decode(), r'INFO: Collection (updated:|saved as)')
        search_key = ('portable_data_hash'
                      if '--portable-data-hash' in extra_args else 'uuid')
        collection_list = arvados.api('v1').collections().list(
            filters=[[search_key, '=', stdout.decode().strip()]]
        ).execute().get('items', [])
        self.assertEqual(1, len(collection_list))
        return collection_list[0]

    def test_all_expired_signatures_invalidates_cache(self):
        self.authorize_with('active')
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'somefile.txt'), 'w') as f:
            f.write('foo')
        # Upload a directory and get the cache file name
        p = subprocess.Popen([sys.executable, arv_put.__file__, tmpdir],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=self.ENVIRON)
        (_, err) = p.communicate()
        self.assertRegex(err.decode(), r'INFO: Creating new cache file at ')
        self.assertEqual(p.returncode, 0)
        cache_filepath = re.search(r'INFO: Creating new cache file at (.*)',
                                   err.decode()).groups()[0]
        self.assertTrue(os.path.isfile(cache_filepath))
        # Load the cache file contents and modify the manifest to simulate
        # an expired access token
        with open(cache_filepath, 'r') as c:
            cache = json.load(c)
        self.assertRegex(cache['manifest'], r'\+A\S+\@')
        a_month_ago = datetime.datetime.now() - datetime.timedelta(days=30)
        cache['manifest'] = re.sub(
            r'\@.*? ',
            "@{} ".format(self.datetime_to_hex(a_month_ago)),
            cache['manifest'])
        with open(cache_filepath, 'w') as c:
            c.write(json.dumps(cache))
        # Re-run the upload and expect to get an invalid cache message
        p = subprocess.Popen([sys.executable, arv_put.__file__, tmpdir],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=self.ENVIRON)
        (_, err) = p.communicate()
        self.assertRegex(
            err.decode(),
            r'INFO: Cache expired, starting from scratch.*')
        self.assertEqual(p.returncode, 0)

    def test_invalid_signature_in_cache(self):
        for batch_mode in [False, True]:
            self.authorize_with('active')
            tmpdir = self.make_tmpdir()
            with open(os.path.join(tmpdir, 'somefile.txt'), 'w') as f:
                f.write('foo')
            # Upload a directory and get the cache file name
            arv_put_args = [tmpdir]
            if batch_mode:
                arv_put_args = ['--batch'] + arv_put_args
            p = subprocess.Popen([sys.executable, arv_put.__file__] + arv_put_args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=self.ENVIRON)
            (_, err) = p.communicate()
            self.assertRegex(err.decode(), r'INFO: Creating new cache file at ')
            self.assertEqual(p.returncode, 0)
            cache_filepath = re.search(r'INFO: Creating new cache file at (.*)',
                                    err.decode()).groups()[0]
            self.assertTrue(os.path.isfile(cache_filepath))
            # Load the cache file contents and modify the manifest to simulate
            # an invalid access token
            with open(cache_filepath, 'r') as c:
                cache = json.load(c)
            self.assertRegex(cache['manifest'], r'\+A\S+\@')
            cache['manifest'] = re.sub(
                r'\+A.*\@',
                "+Aabcdef0123456789abcdef0123456789abcdef01@",
                cache['manifest'])
            with open(cache_filepath, 'w') as c:
                c.write(json.dumps(cache))
            # Re-run the upload and expect to get an invalid cache message
            p = subprocess.Popen([sys.executable, arv_put.__file__] + arv_put_args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=self.ENVIRON)
            (_, err) = p.communicate()
            if not batch_mode:
                self.assertRegex(
                    err.decode(),
                    r'ERROR: arv-put: Resume cache contains invalid signature.*')
                self.assertEqual(p.returncode, 1)
            else:
                self.assertRegex(
                    err.decode(),
                    r'Invalid signatures on cache file \'.*\' while being run in \'batch mode\' -- continuing anyways.*')
                self.assertEqual(p.returncode, 0)

    def test_single_expired_signature_reuploads_file(self):
        self.authorize_with('active')
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'foofile.txt'), 'w') as f:
            f.write('foo')
        # Write a second file on its own subdir to force a new stream
        os.mkdir(os.path.join(tmpdir, 'bar'))
        with open(os.path.join(tmpdir, 'bar', 'barfile.txt'), 'w') as f:
            f.write('bar')
        # Upload a directory and get the cache file name
        p = subprocess.Popen([sys.executable, arv_put.__file__, tmpdir],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=self.ENVIRON)
        (_, err) = p.communicate()
        self.assertRegex(err.decode(), r'INFO: Creating new cache file at ')
        self.assertEqual(p.returncode, 0)
        cache_filepath = re.search(r'INFO: Creating new cache file at (.*)',
                                   err.decode()).groups()[0]
        self.assertTrue(os.path.isfile(cache_filepath))
        # Load the cache file contents and modify the manifest to simulate
        # an expired access token
        with open(cache_filepath, 'r') as c:
            cache = json.load(c)
        self.assertRegex(cache['manifest'], r'\+A\S+\@')
        a_month_ago = datetime.datetime.now() - datetime.timedelta(days=30)
        # Make one of the signatures appear to have expired
        cache['manifest'] = re.sub(
            r'\@.*? 3:3:barfile.txt',
            "@{} 3:3:barfile.txt".format(self.datetime_to_hex(a_month_ago)),
            cache['manifest'])
        with open(cache_filepath, 'w') as c:
            c.write(json.dumps(cache))
        # Re-run the upload and expect to get an invalid cache message
        p = subprocess.Popen([sys.executable, arv_put.__file__, tmpdir],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             env=self.ENVIRON)
        (_, err) = p.communicate()
        self.assertRegex(
            err.decode(),
            r'WARNING: Uploaded file \'.*barfile.txt\' access token expired, will re-upload it from scratch')
        self.assertEqual(p.returncode, 0)
        # Confirm that the resulting cache is different from the last run.
        with open(cache_filepath, 'r') as c2:
            new_cache = json.load(c2)
        self.assertNotEqual(cache['manifest'], new_cache['manifest'])

    def test_put_collection_with_later_update(self):
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection("", ['--no-progress', tmpdir])
        self.assertNotEqual(None, col['uuid'])
        # Add a new file to the directory
        with open(os.path.join(tmpdir, 'file2'), 'w') as f:
            f.write('The quick brown fox jumped over the lazy dog')
        updated_col = self.run_and_find_collection("", ['--no-progress', '--update-collection', col['uuid'], tmpdir])
        self.assertEqual(col['uuid'], updated_col['uuid'])
        # Get the manifest and check that the new file is being included
        c = arv_put.api_client.collections().get(uuid=updated_col['uuid']).execute()
        self.assertRegex(c['manifest_text'], r'^\..* .*:44:file2\n')

    def test_put_collection_with_utc_expiring_datetime(self):
        tmpdir = self.make_tmpdir()
        trash_at = (datetime.datetime.utcnow() + datetime.timedelta(days=90)).strftime('%Y%m%dT%H%MZ')
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection(
            "",
            ['--no-progress', '--trash-at', trash_at, tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        self.assertEqual(ciso8601.parse_datetime(trash_at),
            ciso8601.parse_datetime(c['trash_at']))

    def test_put_collection_with_timezone_aware_expiring_datetime(self):
        tmpdir = self.make_tmpdir()
        trash_at = (datetime.datetime.utcnow() + datetime.timedelta(days=90)).strftime('%Y%m%dT%H%M-0300')
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection(
            "",
            ['--no-progress', '--trash-at', trash_at, tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        self.assertEqual(
            ciso8601.parse_datetime(trash_at).replace(tzinfo=None) + datetime.timedelta(hours=3),
            ciso8601.parse_datetime(c['trash_at']).replace(tzinfo=None))

    def test_put_collection_with_timezone_naive_expiring_datetime(self):
        tmpdir = self.make_tmpdir()
        trash_at = (datetime.datetime.utcnow() + datetime.timedelta(days=90)).strftime('%Y%m%dT%H%M')
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection(
            "",
            ['--no-progress', '--trash-at', trash_at, tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        if time.daylight:
            offset = datetime.timedelta(seconds=time.altzone)
        else:
            offset = datetime.timedelta(seconds=time.timezone)
        self.assertEqual(
            ciso8601.parse_datetime(trash_at) + offset,
            ciso8601.parse_datetime(c['trash_at']).replace(tzinfo=None))

    def test_put_collection_with_expiring_date_only(self):
        tmpdir = self.make_tmpdir()
        trash_at = '2140-01-01'
        end_of_day = datetime.timedelta(hours=23, minutes=59, seconds=59)
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection(
            "",
            ['--no-progress', '--trash-at', trash_at, tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        if time.daylight:
            offset = datetime.timedelta(seconds=time.altzone)
        else:
            offset = datetime.timedelta(seconds=time.timezone)
        self.assertEqual(
            ciso8601.parse_datetime(trash_at) + end_of_day + offset,
            ciso8601.parse_datetime(c['trash_at']).replace(tzinfo=None))

    def test_put_collection_with_invalid_absolute_expiring_datetimes(self):
        cases = ['2100', '210010','2100-10', '2100-Oct']
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        for test_datetime in cases:
            with self.assertRaises(AssertionError):
                self.run_and_find_collection(
                    "",
                    ['--no-progress', '--trash-at', test_datetime, tmpdir])

    def test_put_collection_with_relative_expiring_datetime(self):
        expire_after = 7
        dt_before = datetime.datetime.utcnow() + datetime.timedelta(days=expire_after)
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        col = self.run_and_find_collection(
            "",
            ['--no-progress', '--trash-after', str(expire_after), tmpdir])
        self.assertNotEqual(None, col['uuid'])
        dt_after = datetime.datetime.utcnow() + datetime.timedelta(days=expire_after)
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        trash_at = ciso8601.parse_datetime(c['trash_at']).replace(tzinfo=None)
        self.assertTrue(dt_before < trash_at)
        self.assertTrue(dt_after > trash_at)

    def test_put_collection_with_invalid_relative_expiring_datetime(self):
        expire_after = 0 # Must be >= 1
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'file1'), 'w') as f:
            f.write('Relaxing in basins at the end of inlets terminates the endless tests from the box')
        with self.assertRaises(AssertionError):
            self.run_and_find_collection(
                "",
                ['--no-progress', '--trash-after', str(expire_after), tmpdir])

    def test_upload_directory_reference_without_trailing_slash(self):
        tmpdir1 = self.make_tmpdir()
        tmpdir2 = self.make_tmpdir()
        with open(os.path.join(tmpdir1, 'foo'), 'w') as f:
            f.write('This is foo')
        with open(os.path.join(tmpdir2, 'bar'), 'w') as f:
            f.write('This is not foo')
        # Upload one directory and one file
        col = self.run_and_find_collection("", ['--no-progress',
                                                tmpdir1,
                                                os.path.join(tmpdir2, 'bar')])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        # Check that 'foo' was written inside a subcollection
        # OTOH, 'bar' should have been directly uploaded on the root collection
        self.assertRegex(c['manifest_text'], r'^\. .*:15:bar\n\./.+ .*:11:foo\n')

    def test_upload_directory_reference_with_trailing_slash(self):
        tmpdir1 = self.make_tmpdir()
        tmpdir2 = self.make_tmpdir()
        with open(os.path.join(tmpdir1, 'foo'), 'w') as f:
            f.write('This is foo')
        with open(os.path.join(tmpdir2, 'bar'), 'w') as f:
            f.write('This is not foo')
        # Upload one directory (with trailing slash) and one file
        col = self.run_and_find_collection("", ['--no-progress',
                                                tmpdir1 + os.sep,
                                                os.path.join(tmpdir2, 'bar')])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        # Check that 'foo' and 'bar' were written at the same level
        self.assertRegex(c['manifest_text'], r'^\. .*:15:bar .*:11:foo\n')

    def test_put_collection_with_high_redundancy(self):
        # Write empty data: we're not testing CollectionWriter, just
        # making sure collections.create tells the API server what our
        # desired replication level is.
        collection = self.run_and_find_collection("", ['--replication', '4'])
        self.assertEqual(4, collection['replication_desired'])

    def test_put_collection_with_default_redundancy(self):
        collection = self.run_and_find_collection("")
        self.assertEqual(None, collection['replication_desired'])

    def test_put_collection_with_unnamed_project_link(self):
        link = self.run_and_find_collection(
            "Test unnamed collection",
            ['--portable-data-hash', '--project-uuid', self.PROJECT_UUID])
        username = pwd.getpwuid(os.getuid()).pw_name
        self.assertRegex(
            link['name'],
            r'^Saved at .* by {}@'.format(re.escape(username)))

    def test_put_collection_with_name_and_no_project(self):
        link_name = 'Test Collection Link in home project'
        collection = self.run_and_find_collection(
            "Test named collection in home project",
            ['--portable-data-hash', '--name', link_name])
        self.assertEqual(link_name, collection['name'])
        my_user_uuid = self.current_user()['uuid']
        self.assertEqual(my_user_uuid, collection['owner_uuid'])

    def test_put_collection_with_named_project_link(self):
        link_name = 'Test auto Collection Link'
        collection = self.run_and_find_collection("Test named collection",
                                      ['--portable-data-hash',
                                       '--name', link_name,
                                       '--project-uuid', self.PROJECT_UUID])
        self.assertEqual(link_name, collection['name'])

    def test_put_collection_with_storage_classes_specified(self):
        collection = self.run_and_find_collection("", ['--storage-classes', 'hot'])
        self.assertEqual(len(collection['storage_classes_desired']), 1)
        self.assertEqual(collection['storage_classes_desired'][0], 'hot')

    def test_put_collection_with_multiple_storage_classes_specified(self):
        collection = self.run_and_find_collection("", ['--storage-classes', ' foo, bar  ,baz'])
        self.assertEqual(len(collection['storage_classes_desired']), 3)
        self.assertEqual(collection['storage_classes_desired'], ['foo', 'bar', 'baz'])

    def test_put_collection_without_storage_classes_specified(self):
        collection = self.run_and_find_collection("")
        self.assertEqual(len(collection['storage_classes_desired']), 1)
        self.assertEqual(collection['storage_classes_desired'][0], 'default')

    def test_exclude_filename_pattern(self):
        tmpdir = self.make_tmpdir()
        tmpsubdir = os.path.join(tmpdir, 'subdir')
        os.mkdir(tmpsubdir)
        for fname in ['file1', 'file2', 'file3']:
            with open(os.path.join(tmpdir, "%s.txt" % fname), 'w') as f:
                f.write("This is %s" % fname)
            with open(os.path.join(tmpsubdir, "%s.txt" % fname), 'w') as f:
                f.write("This is %s" % fname)
        col = self.run_and_find_collection("", ['--no-progress',
                                                '--exclude', '*2.txt',
                                                '--exclude', 'file3.*',
                                                 tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        # None of the file2.txt & file3.txt should have been uploaded
        self.assertRegex(c['manifest_text'], r'^.*:file1.txt')
        self.assertNotRegex(c['manifest_text'], r'^.*:file2.txt')
        self.assertNotRegex(c['manifest_text'], r'^.*:file3.txt')

    def test_exclude_filepath_pattern(self):
        tmpdir = self.make_tmpdir()
        tmpsubdir = os.path.join(tmpdir, 'subdir')
        os.mkdir(tmpsubdir)
        for fname in ['file1', 'file2', 'file3']:
            with open(os.path.join(tmpdir, "%s.txt" % fname), 'w') as f:
                f.write("This is %s" % fname)
            with open(os.path.join(tmpsubdir, "%s.txt" % fname), 'w') as f:
                f.write("This is %s" % fname)
        col = self.run_and_find_collection("", ['--no-progress',
                                                '--exclude', 'subdir/*2.txt',
                                                '--exclude', './file1.*',
                                                 tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        # Only tmpdir/file1.txt & tmpdir/subdir/file2.txt should have been excluded
        self.assertNotRegex(c['manifest_text'],
                            r'^\./%s.*:file1.txt' % os.path.basename(tmpdir))
        self.assertNotRegex(c['manifest_text'],
                            r'^\./%s/subdir.*:file2.txt' % os.path.basename(tmpdir))
        self.assertRegex(c['manifest_text'],
                         r'^\./%s.*:file2.txt' % os.path.basename(tmpdir))
        self.assertRegex(c['manifest_text'], r'^.*:file3.txt')

    def test_unicode_on_filename(self):
        tmpdir = self.make_tmpdir()
        fname = u"i❤arvados.txt"
        with open(os.path.join(tmpdir, fname), 'w') as f:
            f.write("This is a unicode named file")
        col = self.run_and_find_collection("", ['--no-progress', tmpdir])
        self.assertNotEqual(None, col['uuid'])
        c = arv_put.api_client.collections().get(uuid=col['uuid']).execute()
        self.assertTrue(fname in c['manifest_text'], u"{} does not include {}".format(c['manifest_text'], fname))

    def test_silent_mode_no_errors(self):
        self.authorize_with('active')
        tmpdir = self.make_tmpdir()
        with open(os.path.join(tmpdir, 'test.txt'), 'w') as f:
            f.write('hello world')
        pipe = subprocess.Popen(
            [sys.executable, arv_put.__file__] + ['--silent', tmpdir],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, env=self.ENVIRON)
        stdout, stderr = pipe.communicate()
        # No console output should occur on normal operations
        self.assertNotRegex(stderr.decode(), r'.+')
        self.assertNotRegex(stdout.decode(), r'.+')

    def test_silent_mode_does_not_avoid_error_messages(self):
        self.authorize_with('active')
        pipe = subprocess.Popen(
            [sys.executable, arv_put.__file__] + ['--silent',
                                                  '/path/not/existant'],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, env=self.ENVIRON)
        stdout, stderr = pipe.communicate()
        # Error message should be displayed when errors happen
        self.assertRegex(stderr.decode(), r'.*ERROR:.*')
        self.assertNotRegex(stdout.decode(), r'.+')


if __name__ == '__main__':
    unittest.main()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
from builtins import object
import bz2
import gzip
import io
import mock
import os
import unittest
import hashlib

import arvados
from arvados import StreamReader, StreamFileReader
from arvados._ranges import Range

from . import arvados_testutil as tutil
from . import run_test_server

class StreamFileReaderTestCase(unittest.TestCase):
    def make_count_reader(self):
        stream = tutil.MockStreamReader('.', '01234', '34567', '67890')
        return StreamFileReader(stream, [Range(1, 0, 3), Range(6, 3, 3), Range(11, 6, 3)],
                                'count.txt')

    def test_read_block_crossing_behavior(self):
        # read() calls will be aligned on block boundaries - see #3663.
        sfile = self.make_count_reader()
        self.assertEqual(b'123', sfile.read(10))

    def test_small_read(self):
        sfile = self.make_count_reader()
        self.assertEqual(b'12', sfile.read(2))

    def test_successive_reads(self):
        sfile = self.make_count_reader()
        for expect in [b'123', b'456', b'789', b'']:
            self.assertEqual(expect, sfile.read(10))

    def test_readfrom_spans_blocks(self):
        sfile = self.make_count_reader()
        self.assertEqual(b'6789', sfile.readfrom(5, 12))

    def test_small_readfrom_spanning_blocks(self):
        sfile = self.make_count_reader()
        self.assertEqual(b'2345', sfile.readfrom(1, 4))

    def test_readall(self):
        sfile = self.make_count_reader()
        self.assertEqual(b'123456789', b''.join(sfile.readall()))

    def test_one_arg_seek(self):
        self.test_absolute_seek([])

    def test_absolute_seek(self, args=[os.SEEK_SET]):
        sfile = self.make_count_reader()
        sfile.seek(6, *args)
        self.assertEqual(b'78', sfile.read(2))
        sfile.seek(4, *args)
        self.assertEqual(b'56', sfile.read(2))

    def test_relative_seek(self, args=[os.SEEK_CUR]):
        sfile = self.make_count_reader()
        self.assertEqual(b'12', sfile.read(2))
        sfile.seek(2, *args)
        self.assertEqual(b'56', sfile.read(2))

    def test_end_seek(self):
        sfile = self.make_count_reader()
        sfile.seek(-6, os.SEEK_END)
        self.assertEqual(b'45', sfile.read(2))

    def test_seek_min_zero(self):
        sfile = self.make_count_reader()
        self.assertEqual(0, sfile.tell())
        with self.assertRaises(IOError):
            sfile.seek(-2, os.SEEK_SET)
        self.assertEqual(0, sfile.tell())

    def test_seek_max_size(self):
        sfile = self.make_count_reader()
        sfile.seek(2, os.SEEK_END)
        # POSIX permits seeking past end of file.
        self.assertEqual(11, sfile.tell())

    def test_size(self):
        self.assertEqual(9, self.make_count_reader().size())

    def test_tell_after_block_read(self):
        sfile = self.make_count_reader()
        sfile.read(5)
        self.assertEqual(3, sfile.tell())

    def test_tell_after_small_read(self):
        sfile = self.make_count_reader()
        sfile.read(1)
        self.assertEqual(1, sfile.tell())

    def test_no_read_after_close(self):
        sfile = self.make_count_reader()
        sfile.close()
        self.assertRaises(ValueError, sfile.read, 2)

    def test_context(self):
        with self.make_count_reader() as sfile:
            self.assertFalse(sfile.closed, "reader is closed inside context")
            self.assertEqual(b'12', sfile.read(2))
        self.assertTrue(sfile.closed, "reader is open after context")

    def make_newlines_reader(self):
        stream = tutil.MockStreamReader('.', 'one\ntwo\n\nth', 'ree\nfour\n\n')
        return StreamFileReader(stream, [Range(0, 0, 11), Range(11, 11, 10)], 'count.txt')

    def check_lines(self, actual):
        self.assertEqual(['one\n', 'two\n', '\n', 'three\n', 'four\n', '\n'],
                         actual)

    def test_readline(self):
        reader = self.make_newlines_reader()
        actual = []
        while True:
            data = reader.readline()
            if not data:
                break
            actual.append(data)
        self.check_lines(actual)

    def test_readlines(self):
        self.check_lines(self.make_newlines_reader().readlines())

    def test_iteration(self):
        self.check_lines(list(iter(self.make_newlines_reader())))

    def test_readline_size(self):
        reader = self.make_newlines_reader()
        self.assertEqual('on', reader.readline(2))
        self.assertEqual('e\n', reader.readline(4))
        self.assertEqual('two\n', reader.readline(6))
        self.assertEqual('\n', reader.readline(8))
        self.assertEqual('thre', reader.readline(4))

    def test_readlines_sizehint(self):
        result = self.make_newlines_reader().readlines(8)
        self.assertEqual(['one\n', 'two\n'], result[:2])
        self.assertNotIn('three\n', result)

    def test_name_attribute(self):
        # Test both .name and .name() (for backward compatibility)
        stream = tutil.MockStreamReader()
        sfile = StreamFileReader(stream, [Range(0, 0, 0)], 'nametest')
        self.assertEqual('nametest', sfile.name)
        self.assertEqual('nametest', sfile.name())

    def check_decompressed_name(self, filename, expect):
        stream = tutil.MockStreamReader('.', '')
        reader = StreamFileReader(stream, [Range(0, 0, 0)], filename)
        self.assertEqual(expect, reader.decompressed_name())

    def test_decompressed_name_uncompressed_file(self):
        self.check_decompressed_name('test.log', 'test.log')

    def test_decompressed_name_gzip_file(self):
        self.check_decompressed_name('test.log.gz', 'test.log')

    def test_decompressed_name_bz2_file(self):
        self.check_decompressed_name('test.log.bz2', 'test.log')

    def check_decompression(self, compress_ext, compress_func):
        test_text = b'decompression\ntest\n'
        test_data = compress_func(test_text)
        stream = tutil.MockStreamReader('.', test_data)
        reader = StreamFileReader(stream, [Range(0, 0, len(test_data))],
                                  'test.' + compress_ext)
        self.assertEqual(test_text, b''.join(reader.readall_decompressed()))

    @staticmethod
    def gzip_compress(data):
        compressed_data = io.BytesIO()
        with gzip.GzipFile(fileobj=compressed_data, mode='wb') as gzip_file:
            gzip_file.write(data)
        return compressed_data.getvalue()

    def test_no_decompression(self):
        self.check_decompression('log', lambda s: s)

    def test_gzip_decompression(self):
        self.check_decompression('gz', self.gzip_compress)

    def test_bz2_decompression(self):
        self.check_decompression('bz2', bz2.compress)

    def test_readline_then_readlines(self):
        reader = self.make_newlines_reader()
        data = reader.readline()
        self.assertEqual('one\n', data)
        data = reader.readlines()
        self.assertEqual(['two\n', '\n', 'three\n', 'four\n', '\n'], data)

    def test_readline_then_readall(self):
        reader = self.make_newlines_reader()
        data = reader.readline()
        self.assertEqual('one\n', data)
        self.assertEqual(b''.join([b'two\n', b'\n', b'three\n', b'four\n', b'\n']), b''.join(reader.readall()))


class StreamRetryTestMixin(object):
    # Define reader_for(coll_name, **kwargs)
    # and read_for_test(reader, size, **kwargs).
    API_COLLECTIONS = run_test_server.fixture('collections')

    def keep_client(self):
        return arvados.KeepClient(proxy='http://[%s]:1' % (tutil.TEST_HOST,),
                                  local_store='')

    def manifest_for(self, coll_name):
        return self.API_COLLECTIONS[coll_name]['manifest_text']

    @tutil.skip_sleep
    def test_success_without_retries(self):
        with tutil.mock_keep_responses('bar', 200):
            reader = self.reader_for('bar_file')
            self.assertEqual(b'bar', self.read_for_test(reader, 3))

    @tutil.skip_sleep
    def test_read_no_default_retry(self):
        with tutil.mock_keep_responses('', 500):
            reader = self.reader_for('user_agreement')
            with self.assertRaises(arvados.errors.KeepReadError):
                self.read_for_test(reader, 10)

    @tutil.skip_sleep
    def test_read_with_instance_retries(self):
        with tutil.mock_keep_responses('foo', 500, 200):
            reader = self.reader_for('foo_file', num_retries=3)
            self.assertEqual(b'foo', self.read_for_test(reader, 3))

    @tutil.skip_sleep
    def test_read_with_method_retries(self):
        with tutil.mock_keep_responses('foo', 500, 200):
            reader = self.reader_for('foo_file')
            self.assertEqual(b'foo',
                             self.read_for_test(reader, 3, num_retries=3))

    @tutil.skip_sleep
    def test_read_instance_retries_exhausted(self):
        with tutil.mock_keep_responses('bar', 500, 500, 500, 500, 200):
            reader = self.reader_for('bar_file', num_retries=3)
            with self.assertRaises(arvados.errors.KeepReadError):
                self.read_for_test(reader, 3)

    @tutil.skip_sleep
    def test_read_method_retries_exhausted(self):
        with tutil.mock_keep_responses('bar', 500, 500, 500, 500, 200):
            reader = self.reader_for('bar_file')
            with self.assertRaises(arvados.errors.KeepReadError):
                self.read_for_test(reader, 3, num_retries=3)

    @tutil.skip_sleep
    def test_method_retries_take_precedence(self):
        with tutil.mock_keep_responses('', 500, 500, 500, 200):
            reader = self.reader_for('user_agreement', num_retries=10)
            with self.assertRaises(arvados.errors.KeepReadError):
                self.read_for_test(reader, 10, num_retries=1)


class StreamReaderTestCase(unittest.TestCase, StreamRetryTestMixin):
    def reader_for(self, coll_name, **kwargs):
        return StreamReader(self.manifest_for(coll_name).split(),
                            self.keep_client(), **kwargs)

    def read_for_test(self, reader, byte_count, **kwargs):
        return reader.readfrom(0, byte_count, **kwargs)

    def test_manifest_text_without_keep_client(self):
        mtext = self.manifest_for('multilevel_collection_1')
        for line in mtext.rstrip('\n').split('\n'):
            reader = StreamReader(line.split())
            self.assertEqual(line + '\n', reader.manifest_text())


class StreamFileReadTestCase(unittest.TestCase, StreamRetryTestMixin):
    def reader_for(self, coll_name, **kwargs):
        return StreamReader(self.manifest_for(coll_name).split(),
                            self.keep_client(), **kwargs).all_files()[0]

    def read_for_test(self, reader, byte_count, **kwargs):
        return reader.read(byte_count, **kwargs)


class StreamFileReadFromTestCase(StreamFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return reader.readfrom(0, byte_count, **kwargs)


class StreamFileReadAllTestCase(StreamFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return b''.join(reader.readall(**kwargs))


class StreamFileReadAllDecompressedTestCase(StreamFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return b''.join(reader.readall_decompressed(**kwargs))


class StreamFileReadlinesTestCase(StreamFileReadTestCase):
    def read_for_test(self, reader, byte_count, **kwargs):
        return ''.join(reader.readlines(**kwargs)).encode()

if __name__ == '__main__':
    unittest.main()



# Copyright (C) The Arvados Authors. All rights reserved.
# Copyright (C) 2018 Genome Research Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import absolute_import
from builtins import range
from past.builtins import basestring
from builtins import object
import arvados
import arvados.commands.ws as ws
import argparse
import json
import re
import os
import stat
from . import put
import time
import subprocess
import logging
import sys
import errno
import arvados.commands._util as arv_cmd
import arvados.collection
import arvados.config as config

from arvados._version import __version__

logger = logging.getLogger('arvados.arv-run')
logger.setLevel(logging.INFO)

class ArvFile(object):
    def __init__(self, prefix, fn):
        self.prefix = prefix
        self.fn = fn

    def __hash__(self):
        return (self.prefix+self.fn).__hash__()

    def __eq__(self, other):
        return (self.prefix == other.prefix) and (self.fn == other.fn)

class UploadFile(ArvFile):
    pass

# Determine if a file is in a collection, and return a tuple consisting of the
# portable data hash and the path relative to the root of the collection.
# Return None if the path isn't with an arv-mount collection or there was is error.
def is_in_collection(root, branch):
    try:
        if root == "/":
            return (None, None)
        fn = os.path.join(root, ".arvados#collection")
        if os.path.exists(fn):
            with open(fn, 'r') as f:
                c = json.load(f)
            return (c["portable_data_hash"], branch)
        else:
            sp = os.path.split(root)
            return is_in_collection(sp[0], os.path.join(sp[1], branch))
    except (IOError, OSError):
        return (None, None)

# Determine the project to place the output of this command by searching upward
# for arv-mount psuedofile indicating the project.  If the cwd isn't within
# an arv-mount project or there is an error, return current_user.
def determine_project(root, current_user):
    try:
        if root == "/":
            return current_user
        fn = os.path.join(root, ".arvados#project")
        if os.path.exists(fn):
            with file(fn, 'r') as f:
                c = json.load(f)
            if 'writable_by' in c and current_user in c['writable_by']:
                return c["uuid"]
            else:
                return current_user
        else:
            sp = os.path.split(root)
            return determine_project(sp[0], current_user)
    except (IOError, OSError):
        return current_user

# Determine if string corresponds to a file, and if that file is part of a
# arv-mounted collection or only local to the machine.  Returns one of
# ArvFile() (file already exists in a collection), UploadFile() (file needs to
# be uploaded to a collection), or simply returns prefix+fn (which yields the
# original parameter string).
def statfile(prefix, fn, fnPattern="$(file %s/%s)", dirPattern="$(dir %s/%s/)", raiseOSError=False):
    absfn = os.path.abspath(fn)
    try:
        st = os.stat(absfn)
        sp = os.path.split(absfn)
        (pdh, branch) = is_in_collection(sp[0], sp[1])
        if pdh:
            if stat.S_ISREG(st.st_mode):
                return ArvFile(prefix, fnPattern % (pdh, branch))
            elif stat.S_ISDIR(st.st_mode):
                return ArvFile(prefix, dirPattern % (pdh, branch))
            else:
                raise Exception("%s is not a regular file or directory" % absfn)
        else:
            # trim leading '/' for path prefix test later
            return UploadFile(prefix, absfn[1:])
    except OSError as e:
        if e.errno == errno.ENOENT and not raiseOSError:
            pass
        else:
            raise

    return prefix+fn

def write_file(collection, pathprefix, fn, flush=False):
    with open(os.path.join(pathprefix, fn), "rb") as src:
        dst = collection.open(fn, "wb")
        r = src.read(1024*128)
        while r:
            dst.write(r)
            r = src.read(1024*128)
        dst.close(flush=flush)

def uploadfiles(files, api, dry_run=False, num_retries=0,
                project=None,
                fnPattern="$(file %s/%s)",
                name=None,
                collection=None,
                packed=True):
    # Find the smallest path prefix that includes all the files that need to be uploaded.
    # This starts at the root and iteratively removes common parent directory prefixes
    # until all file paths no longer have a common parent.
    if files:
        n = True
        pathprefix = "/"
        while n:
            pathstep = None
            for c in files:
                if pathstep is None:
                    sp = c.fn.split('/')
                    if len(sp) < 2:
                        # no parent directories left
                        n = False
                        break
                    # path step takes next directory
                    pathstep = sp[0] + "/"
                else:
                    # check if pathstep is common prefix for all files
                    if not c.fn.startswith(pathstep):
                        n = False
                        break
            if n:
                # pathstep is common parent directory for all files, so remove the prefix
                # from each path
                pathprefix += pathstep
                for c in files:
                    c.fn = c.fn[len(pathstep):]

        logger.info("Upload local files: \"%s\"", '" "'.join([c.fn for c in files]))

    if dry_run:
        logger.info("$(input) is %s", pathprefix.rstrip('/'))
        pdh = "$(input)"
    else:
        files = sorted(files, key=lambda x: x.fn)
        if collection is None:
            collection = arvados.collection.Collection(api_client=api, num_retries=num_retries)
        prev = ""
        for f in files:
            localpath = os.path.join(pathprefix, f.fn)
            if prev and localpath.startswith(prev+"/"):
                # If this path is inside an already uploaded subdirectory,
                # don't redundantly re-upload it.
                # e.g. we uploaded /tmp/foo and the next file is /tmp/foo/bar
                # skip it because it starts with "/tmp/foo/"
                continue
            prev = localpath
            if os.path.isfile(localpath):
                write_file(collection, pathprefix, f.fn, not packed)
            elif os.path.isdir(localpath):
                for root, dirs, iterfiles in os.walk(localpath):
                    root = root[len(pathprefix):]
                    for src in iterfiles:
                        write_file(collection, pathprefix, os.path.join(root, src), not packed)

        pdh = None
        if len(collection) > 0:
            # non-empty collection
            filters = [["portable_data_hash", "=", collection.portable_data_hash()]]
            name_pdh = "%s (%s)" % (name, collection.portable_data_hash())
            if name:
                filters.append(["name", "=", name_pdh])
            if project:
                filters.append(["owner_uuid", "=", project])

            # do the list / create in a loop with up to 2 tries as we are using `ensure_unique_name=False`
            # and there is a potential race with other workflows that may have created the collection
            # between when we list it and find it does not exist and when we attempt to create it.
            tries = 2
            while pdh is None and tries > 0:
                exists = api.collections().list(filters=filters, limit=1).execute(num_retries=num_retries)

                if exists["items"]:
                    item = exists["items"][0]
                    pdh = item["portable_data_hash"]
                    logger.info("Using collection %s (%s)", pdh, item["uuid"])
                else:
                    try:
                        collection.save_new(name=name_pdh, owner_uuid=project, ensure_unique_name=False)
                        pdh = collection.portable_data_hash()
                        logger.info("Uploaded to %s (%s)", pdh, collection.manifest_locator())
                    except arvados.errors.ApiError as ae:
                        tries -= 1
            if pdh is None:
                # Something weird going on here, probably a collection
                # with a conflicting name but wrong PDH.  We won't
                # able to reuse it but we still need to save our
                # collection, so so save it with unique name.
                logger.info("Name conflict on '%s', existing collection has an unexpected portable data hash", name_pdh)
                collection.save_new(name=name_pdh, owner_uuid=project, ensure_unique_name=True)
                pdh = collection.portable_data_hash()
                logger.info("Uploaded to %s (%s)", pdh, collection.manifest_locator())
        else:
            # empty collection
            pdh = collection.portable_data_hash()
            assert (pdh == config.EMPTY_BLOCK_LOCATOR), "Empty collection portable_data_hash did not have expected locator, was %s" % pdh
            logger.debug("Using empty collection %s", pdh)

    for c in files:
        c.keepref = "%s/%s" % (pdh, c.fn)
        c.fn = fnPattern % (pdh, c.fn)


def main(arguments=None):
    raise Exception("Legacy arv-run removed.")

if __name__ == '__main__':
    main()



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import range
import argparse
import atexit
import errno
import glob
import httplib2
import os
import pipes
import random
import re
import shutil
import signal
import socket
import string
import subprocess
import sys
import tempfile
import time
import unittest
import yaml

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

MY_DIRNAME = os.path.dirname(os.path.realpath(__file__))
if __name__ == '__main__' and os.path.exists(
      os.path.join(MY_DIRNAME, '..', 'arvados', '__init__.py')):
    # We're being launched to support another test suite.
    # Add the Python SDK source to the library path.
    sys.path.insert(1, os.path.dirname(MY_DIRNAME))

import arvados
import arvados.config

ARVADOS_DIR = os.path.realpath(os.path.join(MY_DIRNAME, '../../..'))
SERVICES_SRC_DIR = os.path.join(ARVADOS_DIR, 'services')

# Work around https://bugs.python.org/issue27805, should be no longer
# necessary from sometime in Python 3.8.x
if not os.environ.get('ARVADOS_DEBUG', ''):
    WRITE_MODE = 'a'
else:
    WRITE_MODE = 'w'

if 'GOPATH' in os.environ:
    # Add all GOPATH bin dirs to PATH -- but insert them after the
    # ruby gems bin dir, to ensure "bundle" runs the Ruby bundler
    # command, not the golang.org/x/tools/cmd/bundle command.
    gopaths = os.environ['GOPATH'].split(':')
    addbins = [os.path.join(path, 'bin') for path in gopaths]
    newbins = []
    for path in os.environ['PATH'].split(':'):
        newbins.append(path)
        if os.path.exists(os.path.join(path, 'bundle')):
            newbins += addbins
            addbins = []
    newbins += addbins
    os.environ['PATH'] = ':'.join(newbins)

TEST_TMPDIR = os.path.join(ARVADOS_DIR, 'tmp')
if not os.path.exists(TEST_TMPDIR):
    os.mkdir(TEST_TMPDIR)

my_api_host = None
_cached_config = {}
_cached_db_config = {}
_already_used_port = {}

def find_server_pid(PID_PATH, wait=10):
    now = time.time()
    timeout = now + wait
    good_pid = False
    while (not good_pid) and (now <= timeout):
        time.sleep(0.2)
        try:
            with open(PID_PATH, 'r') as f:
                server_pid = int(f.read())
            good_pid = (os.kill(server_pid, 0) is None)
        except EnvironmentError:
            good_pid = False
        now = time.time()

    if not good_pid:
        return None

    return server_pid

def kill_server_pid(pidfile, wait=10, passenger_root=False):
    # Must re-import modules in order to work during atexit
    import os
    import signal
    import subprocess
    import time

    now = time.time()
    startTERM = now
    deadline = now + wait

    if passenger_root:
        # First try to shut down nicely
        restore_cwd = os.getcwd()
        os.chdir(passenger_root)
        subprocess.call([
            'bundle', 'exec', 'passenger', 'stop', '--pid-file', pidfile])
        os.chdir(restore_cwd)
        # Use up to half of the +wait+ period waiting for "passenger
        # stop" to work. If the process hasn't exited by then, start
        # sending TERM signals.
        startTERM += wait//2

    server_pid = None
    while now <= deadline and server_pid is None:
        try:
            with open(pidfile, 'r') as f:
                server_pid = int(f.read())
        except IOError:
            # No pidfile = nothing to kill.
            return
        except ValueError as error:
            # Pidfile exists, but we can't parse it. Perhaps the
            # server has created the file but hasn't written its PID
            # yet?
            print("Parse error reading pidfile {}: {}".format(pidfile, error),
                  file=sys.stderr)
            time.sleep(0.1)
            now = time.time()

    while now <= deadline:
        try:
            exited, _ = os.waitpid(server_pid, os.WNOHANG)
            if exited > 0:
                _remove_pidfile(pidfile)
                return
        except OSError:
            # already exited, or isn't our child process
            pass
        try:
            if now >= startTERM:
                os.kill(server_pid, signal.SIGTERM)
                print("Sent SIGTERM to {} ({})".format(server_pid, pidfile),
                      file=sys.stderr)
        except OSError as error:
            if error.errno == errno.ESRCH:
                # Thrown by os.getpgid() or os.kill() if the process
                # does not exist, i.e., our work here is done.
                _remove_pidfile(pidfile)
                return
            raise
        time.sleep(0.1)
        now = time.time()

    print("Server PID {} ({}) did not exit, giving up after {}s".
          format(server_pid, pidfile, wait),
          file=sys.stderr)

def _remove_pidfile(pidfile):
    try:
        os.unlink(pidfile)
    except:
        if os.path.lexists(pidfile):
            raise

def find_available_port():
    """Return an IPv4 port number that is not in use right now.

    We assume whoever needs to use the returned port is able to reuse
    a recently used port without waiting for TIME_WAIT (see
    SO_REUSEADDR / SO_REUSEPORT).

    Some opportunity for races here, but it's better than choosing
    something at random and not checking at all. If all of our servers
    (hey Passenger) knew that listening on port 0 was a thing, the OS
    would take care of the races, and this wouldn't be needed at all.
    """

    global _already_used_port
    while True:
        sock = socket.socket()
        sock.bind(('0.0.0.0', 0))
        port = sock.getsockname()[1]
        sock.close()
        if port not in _already_used_port:
            _already_used_port[port] = True
            return port

def _wait_until_port_listens(port, timeout=10, warn=True):
    """Wait for a process to start listening on the given port.

    If nothing listens on the port within the specified timeout (given
    in seconds), print a warning on stderr before returning.
    """
    try:
        subprocess.check_output(['which', 'netstat'])
    except subprocess.CalledProcessError:
        print("WARNING: No `netstat` -- cannot wait for port to listen. "+
              "Sleeping 0.5 and hoping for the best.",
              file=sys.stderr)
        time.sleep(0.5)
        return
    deadline = time.time() + timeout
    while time.time() < deadline:
        if re.search(r'\ntcp.*:'+str(port)+' .* LISTEN *\n', subprocess.check_output(['netstat', '-Wln']).decode()):
            return True
        time.sleep(0.1)
    if warn:
        print(
            "WARNING: Nothing is listening on port {} (waited {} seconds).".
            format(port, timeout),
            file=sys.stderr)
    return False

def _logfilename(label):
    """Set up a labelled log file, and return a path to write logs to.

    Normally, the returned path is {tmpdir}/{label}.log.

    In debug mode, logs are also written to stderr, with [label]
    prepended to each line. The returned path is a FIFO.

    +label+ should contain only alphanumerics: it is also used as part
    of the FIFO filename.

    """
    logfilename = os.path.join(TEST_TMPDIR, label+'.log')
    if not os.environ.get('ARVADOS_DEBUG', ''):
        return logfilename
    fifo = os.path.join(TEST_TMPDIR, label+'.fifo')
    try:
        os.remove(fifo)
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
    os.mkfifo(fifo, 0o700)
    stdbuf = ['stdbuf', '-i0', '-oL', '-eL']
    # open(fifo, 'r') would block waiting for someone to open the fifo
    # for writing, so we need a separate cat process to open it for
    # us.
    cat = subprocess.Popen(
        stdbuf+['cat', fifo],
        stdin=open('/dev/null'),
        stdout=subprocess.PIPE)
    tee = subprocess.Popen(
        stdbuf+['tee', '-a', logfilename],
        stdin=cat.stdout,
        stdout=subprocess.PIPE)
    subprocess.Popen(
        stdbuf+['sed', '-e', 's/^/['+label+'] /'],
        stdin=tee.stdout,
        stdout=sys.stderr)
    return fifo

def run(leave_running_atexit=False):
    """Ensure an API server is running, and ARVADOS_API_* env vars have
    admin credentials for it.

    If ARVADOS_TEST_API_HOST is set, a parent process has started a
    test server for us to use: we just need to reset() it using the
    admin token fixture.

    If a previous call to run() started a new server process, and it
    is still running, we just need to reset() it to fixture state and
    return.

    If neither of those options work out, we'll really start a new
    server.
    """
    global my_api_host

    # Delete cached discovery documents.
    #
    # This will clear cached docs that belong to other processes (like
    # concurrent test suites) even if they're still running. They should
    # be able to tolerate that.
    for fn in glob.glob(os.path.join(
            str(arvados.http_cache('discovery')),
            '*,arvados,v1,rest,*')):
        os.unlink(fn)

    pid_file = _pidfile('api')
    pid_file_ok = find_server_pid(pid_file, 0)

    existing_api_host = os.environ.get('ARVADOS_TEST_API_HOST', my_api_host)
    if existing_api_host and pid_file_ok:
        if existing_api_host == my_api_host:
            try:
                return reset()
            except:
                # Fall through to shutdown-and-start case.
                pass
        else:
            # Server was provided by parent. Can't recover if it's
            # unresettable.
            return reset()

    # Before trying to start up our own server, call stop() to avoid
    # "Phusion Passenger Standalone is already running on PID 12345".
    # (If we've gotten this far, ARVADOS_TEST_API_HOST isn't set, so
    # we know the server is ours to kill.)
    stop(force=True)

    restore_cwd = os.getcwd()
    api_src_dir = os.path.join(SERVICES_SRC_DIR, 'api')
    os.chdir(api_src_dir)

    # Either we haven't started a server of our own yet, or it has
    # died, or we have lost our credentials, or something else is
    # preventing us from calling reset(). Start a new one.

    if not os.path.exists('tmp'):
        os.makedirs('tmp')

    if not os.path.exists('tmp/api'):
        os.makedirs('tmp/api')

    if not os.path.exists('tmp/logs'):
        os.makedirs('tmp/logs')

    # Install the git repository fixtures.
    gitdir = os.path.join(SERVICES_SRC_DIR, 'api', 'tmp', 'git')
    gittarball = os.path.join(SERVICES_SRC_DIR, 'api', 'test', 'test.git.tar')
    if not os.path.isdir(gitdir):
        os.makedirs(gitdir)
    subprocess.check_output(['tar', '-xC', gitdir, '-f', gittarball])

    port = internal_port_from_config("RailsAPI")
    env = os.environ.copy()
    env['RAILS_ENV'] = 'test'
    env['ARVADOS_RAILS_LOG_TO_STDOUT'] = '1'
    env.pop('ARVADOS_WEBSOCKETS', None)
    env.pop('ARVADOS_TEST_API_HOST', None)
    env.pop('ARVADOS_API_HOST', None)
    env.pop('ARVADOS_API_HOST_INSECURE', None)
    env.pop('ARVADOS_API_TOKEN', None)
    logf = open(_logfilename('railsapi'), WRITE_MODE)
    railsapi = subprocess.Popen(
        ['bundle', 'exec',
         'passenger', 'start', '-p{}'.format(port),
         '--pid-file', pid_file,
         '--log-file', '/dev/stdout',
         '--ssl',
         '--ssl-certificate', 'tmp/self-signed.pem',
         '--ssl-certificate-key', 'tmp/self-signed.key'],
        env=env, stdin=open('/dev/null'), stdout=logf, stderr=logf)

    if not leave_running_atexit:
        atexit.register(kill_server_pid, pid_file, passenger_root=api_src_dir)

    my_api_host = "127.0.0.1:"+str(port)
    os.environ['ARVADOS_API_HOST'] = my_api_host

    # Make sure the server has written its pid file and started
    # listening on its TCP port
    _wait_until_port_listens(port)
    find_server_pid(pid_file)

    reset()
    os.chdir(restore_cwd)

def reset():
    """Reset the test server to fixture state.

    This resets the ARVADOS_TEST_API_HOST provided by a parent process
    if any, otherwise the server started by run().

    It also resets ARVADOS_* environment vars to point to the test
    server with admin credentials.
    """
    existing_api_host = os.environ.get('ARVADOS_TEST_API_HOST', my_api_host)
    token = auth_token('admin')
    httpclient = httplib2.Http(ca_certs=os.path.join(
        SERVICES_SRC_DIR, 'api', 'tmp', 'self-signed.pem'))
    httpclient.request(
        'https://{}/database/reset'.format(existing_api_host),
        'POST',
        headers={'Authorization': 'OAuth2 {}'.format(token), 'Connection':'close'})

    os.environ['ARVADOS_API_HOST_INSECURE'] = 'true'
    os.environ['ARVADOS_API_TOKEN'] = token
    os.environ['ARVADOS_API_HOST'] = existing_api_host

def stop(force=False):
    """Stop the API server, if one is running.

    If force==False, kill it only if we started it ourselves. (This
    supports the use case where a Python test suite calls run(), but
    run() just uses the ARVADOS_TEST_API_HOST provided by the parent
    process, and the test suite cleans up after itself by calling
    stop(). In this case the test server provided by the parent
    process should be left alone.)

    If force==True, kill it even if we didn't start it
    ourselves. (This supports the use case in __main__, where "run"
    and "stop" happen in different processes.)
    """
    global my_api_host
    if force or my_api_host is not None:
        kill_server_pid(_pidfile('api'))
        my_api_host = None

def get_config():
    with open(os.environ["ARVADOS_CONFIG"]) as f:
        return yaml.safe_load(f)

def internal_port_from_config(service, idx=0):
    return int(urlparse(
        sorted(list(get_config()["Clusters"]["zzzzz"]["Services"][service]["InternalURLs"].keys()))[idx]).
               netloc.split(":")[1])

def external_port_from_config(service):
    return int(urlparse(get_config()["Clusters"]["zzzzz"]["Services"][service]["ExternalURL"]).netloc.split(":")[1])

def run_controller():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    stop_controller()
    logf = open(_logfilename('controller'), WRITE_MODE)
    port = internal_port_from_config("Controller")
    controller = subprocess.Popen(
        ["arvados-server", "controller"],
        stdin=open('/dev/null'), stdout=logf, stderr=logf, close_fds=True)
    with open(_pidfile('controller'), 'w') as f:
        f.write(str(controller.pid))
    _wait_until_port_listens(port)
    return port

def stop_controller():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('controller'))

def run_ws():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    stop_ws()
    port = internal_port_from_config("Websocket")
    logf = open(_logfilename('ws'), WRITE_MODE)
    ws = subprocess.Popen(
        ["arvados-server", "ws"],
        stdin=open('/dev/null'), stdout=logf, stderr=logf, close_fds=True)
    with open(_pidfile('ws'), 'w') as f:
        f.write(str(ws.pid))
    _wait_until_port_listens(port)
    return port

def stop_ws():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('ws'))

def _start_keep(n, blob_signing=False):
    datadir = os.path.join(TEST_TMPDIR, "keep%d.data"%n)
    if os.path.exists(datadir):
        shutil.rmtree(datadir)
    os.mkdir(datadir)
    port = internal_port_from_config("Keepstore", idx=n)

    # Currently, if there are multiple InternalURLs for a single host,
    # the only way to tell a keepstore process which one it's supposed
    # to listen on is to supply a redacted version of the config, with
    # the other InternalURLs removed.
    conf = os.path.join(TEST_TMPDIR, "keep%d.yaml"%n)
    confdata = get_config()
    confdata['Clusters']['zzzzz']['Services']['Keepstore']['InternalURLs'] = {"http://127.0.0.1:%d"%port: {}}
    confdata['Clusters']['zzzzz']['Collections']['BlobSigning'] = blob_signing
    with open(conf, 'w') as f:
        yaml.safe_dump(confdata, f)
    keep_cmd = ["arvados-server", "keepstore", "-config", conf]

    with open(_logfilename('keep{}'.format(n)), WRITE_MODE) as logf:
        with open('/dev/null') as _stdin:
            child = subprocess.Popen(
                keep_cmd, stdin=_stdin, stdout=logf, stderr=logf, close_fds=True)

    print('child.pid is %d'%child.pid, file=sys.stderr)
    with open(_pidfile('keep{}'.format(n)), 'w') as f:
        f.write(str(child.pid))

    _wait_until_port_listens(port)

    return port

def run_keep(num_servers=2, **kwargs):
    stop_keep(num_servers)

    api = arvados.api(
        version='v1',
        host=os.environ['ARVADOS_API_HOST'],
        token=os.environ['ARVADOS_API_TOKEN'],
        insecure=True)

    for d in api.keep_services().list(filters=[['service_type','=','disk']]).execute()['items']:
        api.keep_services().delete(uuid=d['uuid']).execute()
    for d in api.keep_disks().list().execute()['items']:
        api.keep_disks().delete(uuid=d['uuid']).execute()

    for d in range(0, num_servers):
        port = _start_keep(d, **kwargs)
        svc = api.keep_services().create(body={'keep_service': {
            'uuid': 'zzzzz-bi6l4-keepdisk{:07d}'.format(d),
            'service_host': 'localhost',
            'service_port': port,
            'service_type': 'disk',
            'service_ssl_flag': False,
        }}).execute()
        api.keep_disks().create(body={
            'keep_disk': {'keep_service_uuid': svc['uuid'] }
        }).execute()

    # If keepproxy and/or keep-web is running, send SIGHUP to make
    # them discover the new keepstore services.
    for svc in ('keepproxy', 'keep-web'):
        pidfile = _pidfile('keepproxy')
        if os.path.exists(pidfile):
            try:
                with open(pidfile) as pid:
                    os.kill(int(pid.read()), signal.SIGHUP)
            except OSError:
                os.remove(pidfile)

def _stop_keep(n):
    kill_server_pid(_pidfile('keep{}'.format(n)))

def stop_keep(num_servers=2):
    for n in range(0, num_servers):
        _stop_keep(n)

def run_keep_proxy():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        os.environ["ARVADOS_KEEP_SERVICES"] = "http://localhost:{}".format(internal_port_from_config('Keepproxy'))
        return
    stop_keep_proxy()

    port = internal_port_from_config("Keepproxy")
    env = os.environ.copy()
    env['ARVADOS_API_TOKEN'] = auth_token('anonymous')
    logf = open(_logfilename('keepproxy'), WRITE_MODE)
    kp = subprocess.Popen(
        ['keepproxy'], env=env, stdin=open('/dev/null'), stdout=logf, stderr=logf, close_fds=True)

    with open(_pidfile('keepproxy'), 'w') as f:
        f.write(str(kp.pid))
    _wait_until_port_listens(port)

    print("Using API %s token %s" % (os.environ['ARVADOS_API_HOST'], auth_token('admin')), file=sys.stdout)
    api = arvados.api(
        version='v1',
        host=os.environ['ARVADOS_API_HOST'],
        token=auth_token('admin'),
        insecure=True)
    for d in api.keep_services().list(
            filters=[['service_type','=','proxy']]).execute()['items']:
        api.keep_services().delete(uuid=d['uuid']).execute()
    api.keep_services().create(body={'keep_service': {
        'service_host': 'localhost',
        'service_port': port,
        'service_type': 'proxy',
        'service_ssl_flag': False,
    }}).execute()
    os.environ["ARVADOS_KEEP_SERVICES"] = "http://localhost:{}".format(port)
    _wait_until_port_listens(port)

def stop_keep_proxy():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('keepproxy'))

def run_arv_git_httpd():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    stop_arv_git_httpd()

    gitport = internal_port_from_config("GitHTTP")
    env = os.environ.copy()
    env.pop('ARVADOS_API_TOKEN', None)
    logf = open(_logfilename('arv-git-httpd'), WRITE_MODE)
    agh = subprocess.Popen(['arv-git-httpd'],
        env=env, stdin=open('/dev/null'), stdout=logf, stderr=logf)
    with open(_pidfile('arv-git-httpd'), 'w') as f:
        f.write(str(agh.pid))
    _wait_until_port_listens(gitport)

def stop_arv_git_httpd():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('arv-git-httpd'))

def run_keep_web():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    stop_keep_web()

    keepwebport = internal_port_from_config("WebDAV")
    env = os.environ.copy()
    logf = open(_logfilename('keep-web'), WRITE_MODE)
    keepweb = subprocess.Popen(
        ['keep-web'],
        env=env, stdin=open('/dev/null'), stdout=logf, stderr=logf)
    with open(_pidfile('keep-web'), 'w') as f:
        f.write(str(keepweb.pid))
    _wait_until_port_listens(keepwebport)

def stop_keep_web():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('keep-web'))

def run_nginx():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    stop_nginx()
    nginxconf = {}
    nginxconf['LISTENHOST'] = 'localhost'
    nginxconf['CONTROLLERPORT'] = internal_port_from_config("Controller")
    nginxconf['CONTROLLERSSLPORT'] = external_port_from_config("Controller")
    nginxconf['KEEPWEBPORT'] = internal_port_from_config("WebDAV")
    nginxconf['KEEPWEBDLSSLPORT'] = external_port_from_config("WebDAVDownload")
    nginxconf['KEEPWEBSSLPORT'] = external_port_from_config("WebDAV")
    nginxconf['KEEPPROXYPORT'] = internal_port_from_config("Keepproxy")
    nginxconf['KEEPPROXYSSLPORT'] = external_port_from_config("Keepproxy")
    nginxconf['GITPORT'] = internal_port_from_config("GitHTTP")
    nginxconf['GITSSLPORT'] = external_port_from_config("GitHTTP")
    nginxconf['HEALTHPORT'] = internal_port_from_config("Health")
    nginxconf['HEALTHSSLPORT'] = external_port_from_config("Health")
    nginxconf['WSPORT'] = internal_port_from_config("Websocket")
    nginxconf['WSSSLPORT'] = external_port_from_config("Websocket")
    nginxconf['WORKBENCH1PORT'] = internal_port_from_config("Workbench1")
    nginxconf['WORKBENCH1SSLPORT'] = external_port_from_config("Workbench1")
    nginxconf['SSLCERT'] = os.path.join(SERVICES_SRC_DIR, 'api', 'tmp', 'self-signed.pem')
    nginxconf['SSLKEY'] = os.path.join(SERVICES_SRC_DIR, 'api', 'tmp', 'self-signed.key')
    nginxconf['ACCESSLOG'] = _logfilename('nginx_access')
    nginxconf['ERRORLOG'] = _logfilename('nginx_error')
    nginxconf['TMPDIR'] = TEST_TMPDIR

    conftemplatefile = os.path.join(MY_DIRNAME, 'nginx.conf')
    conffile = os.path.join(TEST_TMPDIR, 'nginx.conf')
    with open(conffile, 'w') as f:
        f.write(re.sub(
            r'{{([A-Z]+[A-Z0-9]+)}}',
            lambda match: str(nginxconf.get(match.group(1))),
            open(conftemplatefile).read()))

    env = os.environ.copy()
    env['PATH'] = env['PATH']+':/sbin:/usr/sbin:/usr/local/sbin'

    nginx = subprocess.Popen(
        ['nginx',
         '-g', 'error_log stderr info;',
         '-g', 'pid '+_pidfile('nginx')+';',
         '-c', conffile],
        env=env, stdin=open('/dev/null'), stdout=sys.stderr)
    _wait_until_port_listens(nginxconf['CONTROLLERSSLPORT'])

def setup_config():
    rails_api_port = find_available_port()
    controller_port = find_available_port()
    controller_external_port = find_available_port()
    websocket_port = find_available_port()
    websocket_external_port = find_available_port()
    workbench1_port = find_available_port()
    workbench1_external_port = find_available_port()
    git_httpd_port = find_available_port()
    git_httpd_external_port = find_available_port()
    health_httpd_port = find_available_port()
    health_httpd_external_port = find_available_port()
    keepproxy_port = find_available_port()
    keepproxy_external_port = find_available_port()
    keepstore_ports = sorted([str(find_available_port()) for _ in range(0,4)])
    keep_web_port = find_available_port()
    keep_web_external_port = find_available_port()
    keep_web_dl_port = find_available_port()
    keep_web_dl_external_port = find_available_port()

    configsrc = os.environ.get("CONFIGSRC", None)
    if configsrc:
        clusterconf = os.path.join(configsrc, "config.yml")
        print("Getting config from %s" % clusterconf, file=sys.stderr)
        pgconnection = yaml.safe_load(open(clusterconf))["Clusters"]["zzzzz"]["PostgreSQL"]["Connection"]
    else:
        # assume "arvados-server install -type test" has set up the
        # conventional db credentials
        pgconnection = {
	    "client_encoding": "utf8",
	    "host": "localhost",
	    "dbname": "arvados_test",
	    "user": "arvados",
	    "password": "insecure_arvados_test",
        }

    localhost = "127.0.0.1"
    services = {
        "RailsAPI": {
            "InternalURLs": {
                "https://%s:%s"%(localhost, rails_api_port): {},
            },
        },
        "Controller": {
            "ExternalURL": "https://%s:%s" % (localhost, controller_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, controller_port): {},
            },
        },
        "Websocket": {
            "ExternalURL": "wss://%s:%s/websocket" % (localhost, websocket_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, websocket_port): {},
            },
        },
        "Workbench1": {
            "ExternalURL": "https://%s:%s/" % (localhost, workbench1_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, workbench1_port): {},
            },
        },
        "GitHTTP": {
            "ExternalURL": "https://%s:%s" % (localhost, git_httpd_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, git_httpd_port): {}
            },
        },
        "Health": {
            "ExternalURL": "https://%s:%s" % (localhost, health_httpd_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, health_httpd_port): {}
            },
        },
        "Keepstore": {
            "InternalURLs": {
                "http://%s:%s"%(localhost, port): {} for port in keepstore_ports
            },
        },
        "Keepproxy": {
            "ExternalURL": "https://%s:%s" % (localhost, keepproxy_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, keepproxy_port): {},
            },
        },
        "WebDAV": {
            "ExternalURL": "https://%s:%s" % (localhost, keep_web_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, keep_web_port): {},
            },
        },
        "WebDAVDownload": {
            "ExternalURL": "https://%s:%s" % (localhost, keep_web_dl_external_port),
            "InternalURLs": {
                "http://%s:%s"%(localhost, keep_web_dl_port): {},
            },
        },
    }

    config = {
        "Clusters": {
            "zzzzz": {
                "ManagementToken": "e687950a23c3a9bceec28c6223a06c79",
                "SystemRootToken": auth_token('system_user'),
                "API": {
                    "RequestTimeout": "30s",
                },
                "Login": {
                    "Test": {
                        "Enable": True,
                        "Users": {
                            "alice": {
                                "Email": "alice@example.com",
                                "Password": "xyzzy"
                            }
                        }
                    },
                },
                "SystemLogs": {
                    "LogLevel": ('info' if os.environ.get('ARVADOS_DEBUG', '') in ['','0'] else 'debug'),
                },
                "PostgreSQL": {
                    "Connection": pgconnection,
                },
                "TLS": {
                    "Insecure": True,
                },
                "Services": services,
                "Users": {
                    "AnonymousUserToken": auth_token('anonymous'),
                    "UserProfileNotificationAddress": "arvados@example.com",
                },
                "Collections": {
                    "CollectionVersioning": True,
                    "BlobSigningKey": "zfhgfenhffzltr9dixws36j1yhksjoll2grmku38mi7yxd66h5j4q9w4jzanezacp8s6q0ro3hxakfye02152hncy6zml2ed0uc",
                    "TrustAllContent": False,
                    "ForwardSlashNameSubstitution": "/",
                    "TrashSweepInterval": "-1s",
                },
                "Git": {
                    "Repositories": os.path.join(SERVICES_SRC_DIR, 'api', 'tmp', 'git', 'test'),
                },
                "Containers": {
                    "JobsAPI": {
                        "GitInternalDir": os.path.join(SERVICES_SRC_DIR, 'api', 'tmp', 'internal.git'),
                    },
                    "SupportedDockerImageFormats": {"v1": {}},
                    "ShellAccess": {
                        "Admin": True,
                        "User": True,
                    },
                },
                "Volumes": {
                    "zzzzz-nyw5e-%015d"%n: {
                        "AccessViaHosts": {
                            "http://%s:%s" % (localhost, keepstore_ports[n]): {},
                        },
                        "Driver": "Directory",
                        "DriverParameters": {
                            "Root": os.path.join(TEST_TMPDIR, "keep%d.data"%n),
                        },
                    } for n in range(len(keepstore_ports))
                },
            },
        },
    }

    conf = os.path.join(TEST_TMPDIR, 'arvados.yml')
    with open(conf, 'w') as f:
        yaml.safe_dump(config, f)

    ex = "export ARVADOS_CONFIG="+conf
    print(ex)


def stop_nginx():
    if 'ARVADOS_TEST_PROXY_SERVICES' in os.environ:
        return
    kill_server_pid(_pidfile('nginx'))

def _pidfile(program):
    return os.path.join(TEST_TMPDIR, program + '.pid')

def fixture(fix):
    '''load a fixture yaml file'''
    with open(os.path.join(SERVICES_SRC_DIR, 'api', "test", "fixtures",
                           fix + ".yml")) as f:
        yaml_file = f.read()
        try:
          trim_index = yaml_file.index("# Test Helper trims the rest of the file")
          yaml_file = yaml_file[0:trim_index]
        except ValueError:
          pass
        return yaml.safe_load(yaml_file)

def auth_token(token_name):
    return fixture("api_client_authorizations")[token_name]["api_token"]

def authorize_with(token_name):
    '''token_name is the symbolic name of the token from the api_client_authorizations fixture'''
    arvados.config.settings()["ARVADOS_API_TOKEN"] = auth_token(token_name)
    arvados.config.settings()["ARVADOS_API_HOST"] = os.environ.get("ARVADOS_API_HOST")
    arvados.config.settings()["ARVADOS_API_HOST_INSECURE"] = "true"

class TestCaseWithServers(unittest.TestCase):
    """TestCase to start and stop supporting Arvados servers.

    Define any of MAIN_SERVER, KEEP_SERVER, and/or KEEP_PROXY_SERVER
    class variables as a dictionary of keyword arguments.  If you do,
    setUpClass will start the corresponding servers by passing these
    keyword arguments to the run, run_keep, and/or run_keep_server
    functions, respectively.  It will also set Arvados environment
    variables to point to these servers appropriately.  If you don't
    run a Keep or Keep proxy server, setUpClass will set up a
    temporary directory for Keep local storage, and set it as
    KEEP_LOCAL_STORE.

    tearDownClass will stop any servers started, and restore the
    original environment.
    """
    MAIN_SERVER = None
    WS_SERVER = None
    KEEP_SERVER = None
    KEEP_PROXY_SERVER = None
    KEEP_WEB_SERVER = None

    @staticmethod
    def _restore_dict(src, dest):
        for key in list(dest.keys()):
            if key not in src:
                del dest[key]
        dest.update(src)

    @classmethod
    def setUpClass(cls):
        cls._orig_environ = os.environ.copy()
        cls._orig_config = arvados.config.settings().copy()
        cls._cleanup_funcs = []
        os.environ.pop('ARVADOS_KEEP_SERVICES', None)
        os.environ.pop('ARVADOS_EXTERNAL_CLIENT', None)
        for server_kwargs, start_func, stop_func in (
                (cls.MAIN_SERVER, run, reset),
                (cls.WS_SERVER, run_ws, stop_ws),
                (cls.KEEP_SERVER, run_keep, stop_keep),
                (cls.KEEP_PROXY_SERVER, run_keep_proxy, stop_keep_proxy),
                (cls.KEEP_WEB_SERVER, run_keep_web, stop_keep_web)):
            if server_kwargs is not None:
                start_func(**server_kwargs)
                cls._cleanup_funcs.append(stop_func)
        if (cls.KEEP_SERVER is None) and (cls.KEEP_PROXY_SERVER is None):
            cls.local_store = tempfile.mkdtemp()
            os.environ['KEEP_LOCAL_STORE'] = cls.local_store
            cls._cleanup_funcs.append(
                lambda: shutil.rmtree(cls.local_store, ignore_errors=True))
        else:
            os.environ.pop('KEEP_LOCAL_STORE', None)
        arvados.config.initialize()

    @classmethod
    def tearDownClass(cls):
        for clean_func in cls._cleanup_funcs:
            clean_func()
        cls._restore_dict(cls._orig_environ, os.environ)
        cls._restore_dict(cls._orig_config, arvados.config.settings())


if __name__ == "__main__":
    actions = [
        'start', 'stop',
        'start_ws', 'stop_ws',
        'start_controller', 'stop_controller',
        'start_keep', 'stop_keep',
        'start_keep_proxy', 'stop_keep_proxy',
        'start_keep-web', 'stop_keep-web',
        'start_arv-git-httpd', 'stop_arv-git-httpd',
        'start_nginx', 'stop_nginx', 'setup_config',
    ]
    parser = argparse.ArgumentParser()
    parser.add_argument('action', type=str, help="one of {}".format(actions))
    parser.add_argument('--auth', type=str, metavar='FIXTURE_NAME', help='Print authorization info for given api_client_authorizations fixture')
    parser.add_argument('--num-keep-servers', metavar='int', type=int, default=2, help="Number of keep servers desired")
    parser.add_argument('--keep-blob-signing', action="store_true", help="Enable blob signing for keepstore servers")

    args = parser.parse_args()

    if args.action not in actions:
        print("Unrecognized action '{}'. Actions are: {}.".
              format(args.action, actions),
              file=sys.stderr)
        sys.exit(1)
    if args.action == 'start':
        stop(force=('ARVADOS_TEST_API_HOST' not in os.environ))
        run(leave_running_atexit=True)
        host = os.environ['ARVADOS_API_HOST']
        if args.auth is not None:
            token = auth_token(args.auth)
            print("export ARVADOS_API_TOKEN={}".format(pipes.quote(token)))
            print("export ARVADOS_API_HOST={}".format(pipes.quote(host)))
            print("export ARVADOS_API_HOST_INSECURE=true")
        else:
            print(host)
    elif args.action == 'stop':
        stop(force=('ARVADOS_TEST_API_HOST' not in os.environ))
    elif args.action == 'start_ws':
        run_ws()
    elif args.action == 'stop_ws':
        stop_ws()
    elif args.action == 'start_controller':
        run_controller()
    elif args.action == 'stop_controller':
        stop_controller()
    elif args.action == 'start_keep':
        run_keep(blob_signing=args.keep_blob_signing, num_servers=args.num_keep_servers)
    elif args.action == 'stop_keep':
        stop_keep(num_servers=args.num_keep_servers)
    elif args.action == 'start_keep_proxy':
        run_keep_proxy()
    elif args.action == 'stop_keep_proxy':
        stop_keep_proxy()
    elif args.action == 'start_arv-git-httpd':
        run_arv_git_httpd()
    elif args.action == 'stop_arv-git-httpd':
        stop_arv_git_httpd()
    elif args.action == 'start_keep-web':
        run_keep_web()
    elif args.action == 'stop_keep-web':
        stop_keep_web()
    elif args.action == 'start_nginx':
        run_nginx()
        print("export ARVADOS_API_HOST=0.0.0.0:{}".format(external_port_from_config('Controller')))
    elif args.action == 'stop_nginx':
        stop_nginx()
    elif args.action == 'setup_config':
        setup_config()
    else:
        raise Exception("action recognized but not implemented!?")



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function
from __future__ import absolute_import
from builtins import range
import unittest

from .performance_profiler import profiled

class PerformanceTestSample(unittest.TestCase):
    def foo(self):
        bar = 64

    @profiled
    def test_profiled_decorator(self):
        j = 0
        for i in range(0,2**20):
            j += i
        self.foo()
        print('Hello')



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

# errors.py - Arvados-specific exceptions.

import json

from apiclient import errors as apiclient_errors
from collections import OrderedDict

class ApiError(apiclient_errors.HttpError):
    def _get_reason(self):
        try:
            return '; '.join(json.loads(self.content.decode('utf-8'))['errors'])
        except (KeyError, TypeError, ValueError):
            return super(ApiError, self)._get_reason()


class KeepRequestError(Exception):
    """Base class for errors accessing Keep services."""
    def __init__(self, message='', request_errors=(), label=""):
        """KeepRequestError(message='', request_errors=(), label="")

        :message:
          A human-readable message describing what Keep operation
          failed.

        :request_errors:
          An iterable that yields 2-tuples of keys (where the key refers to
          some operation that was attempted) to the error encountered when
          talking to it--either an exception, or an HTTP response object.
          These will be packed into an OrderedDict, available through the
          request_errors() method.

        :label:
          A label indicating the type of value in the 'key' position of request_errors.

        """
        self.label = label
        self._request_errors = OrderedDict(request_errors)
        if self._request_errors:
            exc_reports = [self._format_error(*err_pair)
                           for err_pair in self._request_errors.items()]
            base_msg = "{}: {}".format(message, "; ".join(exc_reports))
        else:
            base_msg = message
        super(KeepRequestError, self).__init__(base_msg)
        self.message = message

    def _format_error(self, key, error):
        if isinstance(error, HttpError):
            err_fmt = "{} {} responded with {e.status_code} {e.reason}"
        else:
            err_fmt = "{} {} raised {e.__class__.__name__} ({e})"
        return err_fmt.format(self.label, key, e=error)

    def request_errors(self):
        """request_errors() -> OrderedDict

        The keys of the dictionary are described by `self.label`
        The corresponding value is the exception raised when sending the
        request to it."""
        return self._request_errors


class HttpError(Exception):
    def __init__(self, status_code, reason):
        self.status_code = status_code
        self.reason = reason


class ArgumentError(Exception):
    pass
class SyntaxError(Exception):
    pass
class AssertionError(Exception):
    pass
class CommandFailedError(Exception):
    pass
class KeepReadError(KeepRequestError):
    pass
class KeepWriteError(KeepRequestError):
    pass
class NotFoundError(KeepReadError):
    pass
class NotImplementedError(Exception):
    pass
class NoKeepServersError(Exception):
    pass
class StaleWriterStateError(Exception):
    pass
class FeatureNotEnabledError(Exception):
    pass



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

# Use the "profiled" decorator on a test to get profiling data.
#
# Usage:
#   from performance_profiler import profiled
#
#   # See report in tmp/profile/foobar
#   @profiled
#   def foobar():
#       baz = 1
#
#   See "test_a_sample.py" for a working example.
#
# Performance tests run as part of regular test suite.
# You can also run only the performance tests using one of the following:
#     python -m unittest discover tests.performance
#     ./run-tests.sh WORKSPACE=~/arvados --only sdk/python sdk/python_test="--test-suite=tests.performance"

import functools
import os
import pstats
import sys
import unittest
try:
    import cProfile as profile
except ImportError:
    import profile

output_dir = os.path.abspath(os.path.join('tmp', 'profile'))
if not os.path.exists(output_dir):
    os.makedirs(output_dir)

def profiled(function):
    @functools.wraps(function)
    def profiled_function(*args, **kwargs):
        outfile = open(os.path.join(output_dir, function.__name__), "w")
        caught = None
        pr = profile.Profile()
        pr.enable()
        try:
            return function(*args, **kwargs)
        finally:
            pr.disable()
            ps = pstats.Stats(pr, stream=outfile)
            ps.sort_stats('time').print_stats()
    return profiled_function



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

import subprocess
import time
import os
import re
import sys

SETUP_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_PATHS = {
        SETUP_DIR,
        os.path.abspath(os.path.join(SETUP_DIR, "../../build/version-at-commit.sh"))
        }

def choose_version_from():
    ts = {}
    for path in VERSION_PATHS:
        ts[subprocess.check_output(
            ['git', 'log', '--first-parent', '--max-count=1',
             '--format=format:%ct', path]).strip()] = path

    sorted_ts = sorted(ts.items())
    getver = sorted_ts[-1][1]
    print("Using "+getver+" for version number calculation of "+SETUP_DIR, file=sys.stderr)
    return getver

def git_version_at_commit():
    curdir = choose_version_from()
    myhash = subprocess.check_output(['git', 'log', '-n1', '--first-parent',
                                       '--format=%H', curdir]).strip()
    myversion = subprocess.check_output([SETUP_DIR+'/../../build/version-at-commit.sh', myhash]).strip().decode()
    return myversion

def save_version(setup_dir, module, v):
    v = v.replace("~dev", ".dev").replace("~rc", "rc")
    with open(os.path.join(setup_dir, module, "_version.py"), 'wt') as fp:
        return fp.write("__version__ = '%s'\n" % v)

def read_version(setup_dir, module):
    with open(os.path.join(setup_dir, module, "_version.py"), 'rt') as fp:
        return re.match("__version__ = '(.*)'$", fp.read()).groups()[0]

def get_version(setup_dir, module):
    env_version = os.environ.get("ARVADOS_BUILDING_VERSION")

    if env_version:
        save_version(setup_dir, module, env_version)
    else:
        try:
            save_version(setup_dir, module, git_version_at_commit())
        except (subprocess.CalledProcessError, OSError) as err:
            print("ERROR: {0}".format(err), file=sys.stderr)
            pass

    return read_version(setup_dir, module)

# Called from calculate_python_sdk_cwl_package_versions() in run-library.sh
if __name__ == '__main__':
    print(get_version(SETUP_DIR, "arvados"))



# Copyright (C) The Arvados Authors. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import absolute_import
import arvados
import collections
import copy
import hashlib
import mock
import os
import subprocess
import sys
import tempfile
import unittest
import logging

import arvados.commands.keepdocker as arv_keepdocker
from . import arvados_testutil as tutil
from . import run_test_server


class StopTest(Exception):
    pass


class ArvKeepdockerTestCase(unittest.TestCase, tutil.VersionChecker):
    def run_arv_keepdocker(self, args, err):
        sys.argv = ['arv-keepdocker'] + args
        log_handler = logging.StreamHandler(err)
        arv_keepdocker.logger.addHandler(log_handler)
        try:
            return arv_keepdocker.main()
        finally:
            arv_keepdocker.logger.removeHandler(log_handler)

    def test_unsupported_arg(self):
        out = tutil.StringIO()
        with tutil.redirected_streams(stdout=out, stderr=out), \
             self.assertRaises(SystemExit):
            self.run_arv_keepdocker(['-x=unknown'], sys.stderr)
        self.assertRegex(out.getvalue(), 'unrecognized arguments')

    def test_version_argument(self):
        with tutil.redirected_streams(
                stdout=tutil.StringIO, stderr=tutil.StringIO) as (out, err):
            with self.assertRaises(SystemExit):
                self.run_arv_keepdocker(['--version'], sys.stderr)
        self.assertVersionOutput(out, err)

    @mock.patch('arvados.commands.keepdocker.list_images_in_arv',
                return_value=[])
    @mock.patch('arvados.commands.keepdocker.find_image_hashes',
                return_value=['abc123'])
    @mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                return_value='abc123')
    def test_image_format_compatibility(self, _1, _2, _3):
        old_id = hashlib.sha256(b'old').hexdigest()
        new_id = 'sha256:'+hashlib.sha256(b'new').hexdigest()
        for supported, img_id, expect_ok in [
                (['v1'], old_id, True),
                (['v1'], new_id, False),
                (None, old_id, False),
                ([], old_id, False),
                ([], new_id, False),
                (['v1', 'v2'], new_id, True),
                (['v1'], new_id, False),
                (['v2'], new_id, True)]:

            fakeDD = arvados.api('v1')._rootDesc
            if supported is None:
                del fakeDD['dockerImageFormats']
            else:
                fakeDD['dockerImageFormats'] = supported

            err = tutil.StringIO()
            out = tutil.StringIO()

            with tutil.redirected_streams(stdout=out), \
                 mock.patch('arvados.api') as api, \
                 mock.patch('arvados.commands.keepdocker.popen_docker',
                            return_value=subprocess.Popen(
                                ['echo', img_id],
                                stdout=subprocess.PIPE)), \
                 mock.patch('arvados.commands.keepdocker.prep_image_file',
                            side_effect=StopTest), \
                 self.assertRaises(StopTest if expect_ok else SystemExit):

                api()._rootDesc = fakeDD
                self.run_arv_keepdocker(['--force', 'testimage'], err)

            self.assertEqual(out.getvalue(), '')
            if expect_ok:
                self.assertNotRegex(
                    err.getvalue(), "refusing to store",
                    msg=repr((supported, img_id)))
            else:
                self.assertRegex(
                    err.getvalue(), "refusing to store",
                    msg=repr((supported, img_id)))
            if not supported:
                self.assertRegex(
                    err.getvalue(),
                    "server does not specify supported image formats",
                    msg=repr((supported, img_id)))

        fakeDD = arvados.api('v1')._rootDesc
        fakeDD['dockerImageFormats'] = ['v1']
        err = tutil.StringIO()
        out = tutil.StringIO()
        with tutil.redirected_streams(stdout=out), \
             mock.patch('arvados.api') as api, \
             mock.patch('arvados.commands.keepdocker.popen_docker',
                        return_value=subprocess.Popen(
                            ['echo', new_id],
                            stdout=subprocess.PIPE)), \
             mock.patch('arvados.commands.keepdocker.prep_image_file',
                        side_effect=StopTest), \
             self.assertRaises(StopTest):
            api()._rootDesc = fakeDD
            self.run_arv_keepdocker(
                ['--force', '--force-image-format', 'testimage'], err)
        self.assertRegex(err.getvalue(), "forcing incompatible image")

    def test_tag_given_twice(self):
        with tutil.redirected_streams(stdout=tutil.StringIO, stderr=tutil.StringIO) as (out, err):
            with self.assertRaises(SystemExit):
                self.run_arv_keepdocker(['myrepo:mytag', 'extratag'], sys.stderr)
            self.assertRegex(err.getvalue(), "cannot add tag argument 'extratag'")

    def test_image_given_as_repo_colon_tag(self):
        with self.assertRaises(StopTest), \
             mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                        side_effect=StopTest) as find_image_mock:
            self.run_arv_keepdocker(['repo:tag'], sys.stderr)
        find_image_mock.assert_called_with('repo', 'tag')

        with self.assertRaises(StopTest), \
             mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                        side_effect=StopTest) as find_image_mock:
            self.run_arv_keepdocker(['myreg.example:8888/repo/img:tag'], sys.stderr)
        find_image_mock.assert_called_with('myreg.example:8888/repo/img', 'tag')

    def test_image_has_colons(self):
        with self.assertRaises(StopTest), \
             mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                        side_effect=StopTest) as find_image_mock:
            self.run_arv_keepdocker(['[::1]:8888/repo/img'], sys.stderr)
        find_image_mock.assert_called_with('[::1]:8888/repo/img', 'latest')

        with self.assertRaises(StopTest), \
             mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                        side_effect=StopTest) as find_image_mock:
            self.run_arv_keepdocker(['[::1]/repo/img'], sys.stderr)
        find_image_mock.assert_called_with('[::1]/repo/img', 'latest')

    @mock.patch('arvados.commands.keepdocker.list_images_in_arv',
                return_value=[])
    @mock.patch('arvados.commands.keepdocker.find_image_hashes',
                return_value=['abc123'])
    @mock.patch('arvados.commands.keepdocker.find_one_image_hash',
                return_value='abc123')
    def test_collection_property_update(self, _1, _2, _3):
        image_id = 'sha256:'+hashlib.sha256(b'image').hexdigest()
        fakeDD = arvados.api('v1')._rootDesc
        fakeDD['dockerImageFormats'] = ['v2']

        err = tutil.StringIO()
        out = tutil.StringIO()
        File = collections.namedtuple('File', ['name'])
        mocked_file = File(name='docker_image')
        mocked_collection = {
            'uuid': 'new-collection-uuid',
            'properties': {
                'responsible_person_uuid': 'person_uuid',
            }
        }

        with tutil.redirected_streams(stdout=out), \
             mock.patch('arvados.api') as api, \
             mock.patch('arvados.commands.keepdocker.popen_docker',
                        return_value=subprocess.Popen(
                            ['echo', image_id],
                            stdout=subprocess.PIPE)), \
             mock.patch('arvados.commands.keepdocker.prep_image_file',
                        return_value=(mocked_file, False)), \
             mock.patch('arvados.commands.put.main',
                        return_value='new-collection-uuid'), \
             self.assertRaises(StopTest):

            api()._rootDesc = fakeDD
            api().collections().get().execute.return_value = copy.deepcopy(mocked_collection)
            api().collections().update().execute.side_effect = StopTest
            self.run_arv_keepdocker(['--force', 'testimage'], err)

        updated_properties = mocked_collection['properties']
        updated_properties.update({'docker-image-repo-tag': 'testimage:latest'})
        api().collections().update.assert_called_with(
            uuid=mocked_collection['uuid'],
            body={'properties': updated_properties})
